"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild
Copyright (c) 2015-present Rapptz (https://github.com/Rapptz/discord.py/blob/master/discord/utils.py)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import colorsys
import datetime
import functools
import json
import logging
import os
import pathlib
import struct
import sys
import typing
from collections.abc import Callable, Iterable
from functools import wraps
from typing import TYPE_CHECKING, Any, ForwardRef, Generic, Literal, Self, TypeVar, Union, cast
from urllib.parse import quote


if TYPE_CHECKING:
    from collections.abc import Generator

    from .types_.colours import Colours
    from .types_.options import WaitPredicateT


try:
    import orjson  # type: ignore

    _from_json: Any = orjson.loads  # type: ignore
except ImportError:
    _from_json: Any = json.loads


PY_312 = sys.version_info >= (3, 12)


__all__ = (
    "MISSING",
    "ColorFormatter",
    "ColourFormatter",
    "_from_json",
    "_is_submodule",
    "date_to_datetime_with_z",
    "handle_user_ids",
    "parse_timestamp",
    "setup_logging",
    "url_encode_datetime",
)

T_co = TypeVar("T_co", covariant=True)


class classproperty(Generic[T_co]):
    def __init__(self, fget: Callable[[Any], T_co]) -> None:
        self.fget = fget

    def __get__(self, instance: Any | None, owner: type[Any]) -> T_co:
        return self.fget(owner)

    def __set__(self, instance: Any | None, value: Any) -> None:
        raise AttributeError("cannot set attribute")


def is_docker() -> bool:
    path: pathlib.Path = pathlib.Path("/proc/self/cgroup")

    exists: bool = path.exists()
    is_file: bool = path.is_file()
    return exists or (is_file and any("docker" in line for line in path.open()))


def stream_supports_colour(stream: Any) -> bool:
    is_a_tty = hasattr(stream, "isatty") and stream.isatty()

    # Pycharm and Vscode support colour in their inbuilt editors
    if "PYCHARM_HOSTED" in os.environ or os.environ.get("TERM_PROGRAM") == "vscode":
        return is_a_tty

    if sys.platform != "win32":
        # Docker does not consistently have a tty attached to it
        return is_a_tty or is_docker()

    # ANSICON checks for things like ConEmu
    # WT_SESSION checks if this is Windows Terminal
    return is_a_tty and ("ANSICON" in os.environ or "WT_SESSION" in os.environ)


def stream_supports_rgb(stream: Any) -> bool:
    if not stream_supports_colour(stream):
        return False

    if "COLORTERM" in os.environ:
        return os.environ["COLORTERM"] in ("truecolor", "24bit")

    return False


def parse_timestamp(timestamp: str) -> datetime.datetime:
    """
    Parses a timestamp in ISO8601 format to a datetime object.

    Parameters
    ----------
    timestamp: str
        The ISO8601 timestamp to be parsed.

    Returns
    -------
    datetime.datetime
        The parsed datetime object.
    """
    return datetime.datetime.fromisoformat(timestamp)


class ColourFormatter(logging.Formatter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._handler: logging.StreamHandler[Any] = kwargs.get("handler", logging.StreamHandler())

        self._supports_colour: bool = stream_supports_colour(self._handler.stream)
        self._supports_rgb: bool = stream_supports_rgb(self._handler.stream)

        self._colours: dict[int, str] = {}
        self._RESET: str = "\033[0m"

        if self._supports_rgb:
            self._colours = {
                logging.DEBUG: "\x1b[40;1m",
                logging.INFO: "\x1b[38;2;100;55;215;1m",
                logging.WARNING: "\x1b[38;2;204;189;51;1m",
                logging.ERROR: "\x1b[38;2;161;38;46m",
                logging.CRITICAL: "\x1b[48;2;161;38;46",
            }

        elif self._supports_colour:
            self._colours = {
                logging.DEBUG: "\x1b[40;1m",
                logging.INFO: "\x1b[34;1m",
                logging.WARNING: "\x1b[33;1m",
                logging.ERROR: "\x1b[31m",
                logging.CRITICAL: "\x1b[41",
            }

        self._FORMATS: dict[int, logging.Formatter] = {
            level: logging.Formatter(
                f"\x1b[30;1m%(asctime)s\x1b[0m {colour}%(levelname)-8s\x1b[0m {colour}%(name)s\x1b[0m %(message)s"
            )
            for level, colour in self._colours.items()
        }

    def format(self, record: logging.LogRecord) -> str:
        formatter: logging.Formatter | None = self._FORMATS.get(record.levelno, None)
        if formatter is None:
            formatter = self._FORMATS[logging.DEBUG]

        # Override the traceback to always print in red
        if record.exc_info:
            text = formatter.formatException(record.exc_info)
            record.exc_text = f"\x1b[31m{text}\x1b[0m"

        output = formatter.format(record)

        # Remove the cache layer
        record.exc_text = None
        return output


ColorFormatter = ColourFormatter


def setup_logging(
    *,
    handler: logging.Handler | None = None,
    formatter: logging.Formatter | None = None,
    level: int | None = None,
    root: bool = True,
) -> None:
    """A helper function to setup logging for your application.

    Parameters
    ----------
    handler: :class:`logging.Handler` | None
        An optional :class:`logging.Handler` to use. Defaults to ``None``, which creates a :class:`logging.StreamHandler`
        by default.
    formatter: :class:`logging.Formatter` | None
        An optional :class:`logging.Formatter` to use. Defaults to ``None``, which uses a custom TrueColour
        formatter by default, falling back to standard colour support and finally no colour support if no colour
        is supported.
    level: int | None
        An optional int indicating the level of logging output. Defaults to ``20``, which is ``INFO``.
    root: bool
        An optional bool indicating whether logging should be setup on the root logger. When ``False``, logging will only be
        setup for twitchio. Defaults to ``True``.

    Examples
    --------

    .. code-block:: python3

        import logging

        import twitchio


        LOGGER: logging.Logger = logging.getLogger(__name__)
        twitchio.utils.setup_logging(level=logging.INFO)
        ...

        arg: str = "World!"
        LOGGER.info("Hello %s", arg)
    """
    if level is None:
        level = logging.INFO

    if handler is None:
        handler = logging.StreamHandler()

    if formatter is None:
        if isinstance(handler, logging.StreamHandler) and stream_supports_colour(handler.stream):  # type: ignore
            formatter = ColourFormatter()
        else:
            dt_fmt = "%Y-%m-%d %H:%M:%S"
            formatter = logging.Formatter("[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style="{")

    if root:
        logger = logging.getLogger()
    else:
        library, _, _ = __name__.partition(".")
        logger = logging.getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)  # type: ignore


# Colour stuff...


class Colour:
    """Helper class for working with colours in TwitchIO.

    .. tip::
        There is an alias for this class called `Color`.

    Supported Operations
    --------------------

    +-------------+----------------------------------------+----------------------------------------------------------+
    | Operation   | Usage(s)                               | Description                                              |
    +=============+========================================+==========================================================+
    | __str__     | str(colour), f"{colour}"               | Returns the colour as a hex string.                      |
    +-------------+----------------------------------------+----------------------------------------------------------+
    | __repr__    | repr(colour), f"{colour!r}"            | Returns the colours official representation.             |
    +-------------+----------------------------------------+----------------------------------------------------------+
    | __eq__      | colour1 == colour2                     | Checks if two colours are equal.                         |
    +-------------+----------------------------------------+----------------------------------------------------------+
    | __format__  | f"{colour:html}", f"{colour:rgb}",     | Returns a colour str in the specified format.            |
    |             | f"{colour:hls}"                        |                                                          |
    +-------------+----------------------------------------+----------------------------------------------------------+
    | __int__     | int(colour)                            | Returns the code as an integer.                          |
    +-------------+----------------------------------------+----------------------------------------------------------+

    """

    __slots__ = (
        "_code",
        "_hex",
        "_hex_clean",
        "_hls",
        "_hls_coords",
        "_html",
        "_rgb",
        "_rgb_coords",
    )

    def __init__(self, data: Colours) -> None:
        self._code: int = data["code"]
        self._hex: str = data["hex"]
        self._hex_clean: str = data["hex_clean"]
        self._html: str = data["html"]
        self._rgb: tuple[int, int, int] = data["rgb"]
        self._hls: tuple[float, float, float] = data["hls"]
        self._rgb_coords: tuple[float, ...] = data["rgb_coords"]
        self._hls_coords: tuple[float, ...] = data["hls_coords"]

    @property
    def code(self) -> int:
        """Property returning the colour as an integer.

        E.g. `16768256`
        """
        return self._code

    @property
    def hex(self) -> str:
        """Property returning the colour as a hex string.

        E.g. `"0xFFDD00"`
        """
        return self._hex

    @property
    def hex_clean(self) -> str:
        """Property returning the colour as a hex string without any prefix.

        E.g. `"FFDD00"`
        """
        return self._hex_clean

    @property
    def html(self) -> str:
        """Property returning the colour as a hex string with a `#` prefix.

        E.g. `"#FFDD00"`
        """
        return self._html

    @property
    def rgb(self) -> tuple[int, int, int]:
        """Property returning the colour as an RGB tuple of `int`.

        E.g. `(255, 221, 0)`
        """
        return self._rgb

    @property
    def hls(self) -> tuple[float, float, float]:
        """Property returning the colour as an HLS tuple of `float`.

        This is not the same as :attr:`~twitchio.Colour.hls_coords`.

        E.g. `(52, 50, 100)`
        """
        return self._hls

    @property
    def rgb_coords(self) -> tuple[float, ...]:
        """Property returning the colour as an RGB tuple of `float` coordinates.

        E.g. `(1.0, 0.8666666666666667, 0.0)`
        """
        return self._rgb_coords

    @property
    def hls_coords(self) -> tuple[float, ...]:
        """Property returning the colour as an HLS tuple of `float` coordinates.

        E.g. `(0.14444444444444446, 0.5, 1.0)`
        """
        return self._hls_coords

    def __repr__(self) -> str:
        return f"<Colour code={self._code} hex={self._hex} html={self._html}>"

    def __str__(self) -> str:
        return self._hex

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Colour):
            return NotImplemented

        return self._code == __value._code

    def __hash__(self) -> int:
        return hash(self._code)

    def __format__(self, format_spec: str) -> str:
        if format_spec == "html":
            return self._html

        elif format_spec == "rgb":
            return f"rgb{self._rgb}"

        elif format_spec == "hls":
            return f"hls{self._hls}"

        else:
            return str(self)

    def __int__(self) -> int:
        return self._code

    @classmethod
    def defaults(cls) -> tuple[Self, ...]:
        """Classmethod which returns a :class:`tuple` of :class:`~twitchio.Colour` representing every default chat colour
        on Twitch.

        Returns
        -------
        tuple[:class:`~twitchio.Colour`]
            A tuple of :class:`~twitchio.Colour` representing every default colour.
        """
        all_ = (
            cls.blue(),
            cls.blue_violet(),
            cls.cadet_blue(),
            cls.chocolate(),
            cls.coral(),
            cls.dodger_blue(),
            cls.firebrick(),
            cls.green(),
            cls.golden_rod(),
            cls.hot_pink(),
            cls.orange_red(),
            cls.sea_green(),
            cls.spring_green(),
            cls.yellow_green(),
        )

        return all_

    @classmethod
    def from_hex(cls, hex_: str, /) -> Self:
        """Class method to create a :class:`~twitchio.Colour` object from a hex string.

        A valid hex string can be in either one of the following formats:

        - `"#FFDD00"`
        - `"FFDD00"`
        - `"0xFFDD00"`

        For ints see: [`.from_int`][twitchio.Colour.from_int].

        Example
        -------
        .. code:: python3

            colour = twitchio.Colour.from_hex("#FFDD00")
            print(colour)


        Parameters
        ----------
        hex_: str
            Positional only. The hex string to convert to a :class:`~twitchio.Colour` object.

        Returns
        -------
        twitchio.Colour
            The :class:`~twitchio.Colour` object created from the hex string.
        """
        value: str = hex_.lstrip("#")
        value = value.removeprefix("0x")

        rgb: tuple[int, int, int] = struct.unpack("BBB", bytes.fromhex(value))
        int_: int = int(value, 16)
        html: str = f"#{value}"
        hexa: str = hex(int_)
        rgb_coords: tuple[float, ...] = tuple(c / 255 for c in rgb)

        hue, light, sat = colorsys.rgb_to_hls(*rgb_coords)
        hls: tuple[float, ...] = (round(hue * 360), round(light * 100), round(sat * 100))

        payload: Colours = {
            "code": int_,
            "hex": hexa,
            "hex_clean": value,
            "html": html,
            "rgb": rgb,
            "hls": hls,
            "rgb_coords": rgb_coords,
            "hls_coords": (hue, light, sat),
        }

        return cls(payload)

    @classmethod
    def from_int(cls, code: int, /) -> Self:
        """Class method to create a :class:`~twitchio.Colour` object from a base-10 or base-16 integer.

        A valid integer can be in either one of the following formats:

        - `16768256`
        - `0xFFDD00`

        For hex strings see: :meth:`Colour.from_hex`.

        Example
        -------
        .. code:: python3

            colour = twitchio.Colour.from_int(0xffdd00)
            print(colour)

        Parameters
        ----------
        code: int
            Positional only. The integer to convert to a :class:`~twitchio.Colour` object.

        Returns
        -------
        twitchio.Colour
            The :class:`~twitchio.Colour` object created from the integer.
        """
        hex_: str = f"{code:X}"

        return cls.from_hex(hex_)

    @classmethod
    def red(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `red` colour from the Twitch API.

        This corresponds to hex: `#FF0000`

        Example
        -------
        .. code:: python3

            red_colour = twitchio.Colour.red()

        """
        return cls.from_hex("FF0000")

    @classmethod
    def blue(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `blue` colour from the Twitch API.

        This corresponds to hex: `#0000FF`

        Example
        -------
        .. code:: python3

            blue_colour = twitchio.Colour.blue()

        """
        return cls.from_hex("0000FF")

    @classmethod
    def green(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `green` colour from the Twitch API.

        This corresponds to hex: `#008000`

        Example
        -------
        .. code:: python3

            green_colour = twitchio.Colour.green()

        """
        return cls.from_hex("008000")

    @classmethod
    def firebrick(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `firebrick` colour from the Twitch API.

        This corresponds to hex: `#B22222`

        Example
        -------
        .. code:: python3

            firebrick_colour = twitchio.Colour.firebrick()

        """
        return cls.from_hex("B22222")

    @classmethod
    def coral(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `coral` colour from the Twitch API.

        This corresponds to hex: `#FF7F50`

        Example
        -------
        .. code:: python3

            coral_colour = twitchio.Colour.coral()

        """
        return cls.from_hex("FF7F50")

    @classmethod
    def yellow_green(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `yellow_green` colour from the Twitch API.

        This corresponds to hex: `#9ACD32`

        Example
        -------
        .. code:: python3

            yellow_green_colour = twitchio.Colour.yellow_green()

        """
        return cls.from_hex("9ACD32")

    @classmethod
    def orange_red(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `orange_red` colour from the Twitch API.

        This corresponds to hex: `#FF4500`

        Example
        -------
        .. code:: python3

            orange_red_colour = twitchio.Colour.orange_red()

        """
        return cls.from_hex("FF4500")

    @classmethod
    def sea_green(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `sea_green` colour from the Twitch API.

        This corresponds to hex: `#2E8B57`

        Example
        -------
        .. code:: python3

            sea_green_colour = twitchio.Colour.sea_green()

        """
        return cls.from_hex("2E8B57")

    @classmethod
    def golden_rod(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `golden_rod` colour from the Twitch API.

        This corresponds to hex: `#DAA520`

        Example
        -------
        .. code:: python3

            golden_rod_colour = twitchio.Colour.golden_rod()

        """
        return cls.from_hex("DAA520")

    @classmethod
    def chocolate(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `chocolate` colour from the Twitch API.

        This corresponds to hex: `#D2691E`

        Example
        -------
        .. code:: python3

            chocolate_colour = twitchio.Colour.chocolate()

        """
        return cls.from_hex("D2691E")

    @classmethod
    def cadet_blue(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `cadet_blue` colour from the Twitch API.

        This corresponds to hex: `#5F9EA0`

        Example
        -------
        .. code:: python3

            cadet_blue_colour = twitchio.Colour.cadet_blue()

        """
        return cls.from_hex("5F9EA0")

    @classmethod
    def dodger_blue(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `dodger_blue` colour from the Twitch API.

        This corresponds to hex: `#1E90FF`

        Example
        -------
        .. code:: python3

            dodger_blue_colour = twitchio.Colour.dodger_blue()

        """
        return cls.from_hex("1E90FF")

    @classmethod
    def hot_pink(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `hot_pink` colour from the Twitch API.

        This corresponds to hex: `#FF69B4`

        Example
        -------
        .. code:: python3

            hot_pink_colour = twitchio.Colour.hot_pink()

        """
        return cls.from_hex("FF69B4")

    @classmethod
    def blue_violet(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `blue_violet` colour from the Twitch API.

        This corresponds to hex: `#8A2BE2`

        Example
        -------
        .. code:: python3

            blue_violet_colour = twitchio.Colour.blue_violet()

        """
        return cls.from_hex("8A2BE2")

    @classmethod
    def spring_green(cls) -> Self:
        """Class method to create a :class:`~twitchio.Colour` with the default `spring_green` colour from the Twitch API.

        This corresponds to hex: `#00FF7F`

        Example
        -------
        .. code:: python3

            sping_green_colour = twitchio.Colour.spring_green()

        """
        return cls.from_hex("00FF7F")

    @classmethod
    def announcement_blue(cls) -> str:
        """A helper class method returning the string "blue", which can be used when sending an announcement.

        Example
        -------

        .. code:: python3

            await ctx.send_announcement("Hello", colour=Colour.announcement_blue())
        """
        return "blue"

    @classmethod
    def announcement_green(cls) -> str:
        """A helper class method returning the string "green", which can be used when sending an announcement.

        Example
        -------

        .. code:: python3

            await ctx.send_announcement("Hello", colour=Colour.announcement_green())
        """
        return "green"

    @classmethod
    def announcement_orange(cls) -> str:
        """A helper class method returning the string "orange", which can be used when sending an announcement.

        Example
        -------

        .. code:: python3

            await ctx.send_announcement("Hello", colour=Colour.announcement_orange())
        """
        return "orange"

    @classmethod
    def announcement_purple(cls) -> str:
        """A helper class method returning the string "purple", which can be used when sending an announcement.

        Example
        -------

        .. code:: python3

            await ctx.send_announcement("Hello", colour=Colour.announcement_purple())
        """
        return "purple"

    @classmethod
    def announcement_primary(cls) -> str:
        """A helper class method returning the string "primary", which can be used when sending an announcement.

        .. note::

            "primary" is the default value used by Twitch and isn't required to be sent.
            The channel's accent color is used to highlight the announcement instead.

        Example
        -------

        .. code:: python3

            await ctx.send_announcement("Hello", colour=Colour.announcement_primary())
        """
        return "primary"


Color = Colour


def chunk_list(sequence: list[Any], n: int) -> Generator[Any, Any, Any]:
    for i in range(0, len(sequence), n):
        yield sequence[i : i + n]


def url_encode_datetime(dt: datetime.datetime) -> str:
    """
    Formats a datetime object to an RFC 3339 compliant string and URL-encodes it.
    If the datetime object does not have a timezone, it is converted to UTC first.

    Parameters
    ----------
    dt : datetime.datetime
        Datetime object.

    Returns
    -------
    str
        The URL encoded parsed datetime object.
    """
    formatted_dt = dt.replace(tzinfo=datetime.UTC).isoformat() if dt.tzinfo is None else dt.isoformat()

    return quote(formatted_dt)


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> str:
        return "..."


MISSING: Any = _MissingSentinel()


class EventWaiter:
    _set: set[Self]

    def __init__(self, *, event: str, predicate: WaitPredicateT | None = None, timeout: float | None = None) -> None:
        self._event: str = event
        self._timeout: float | None = timeout
        self._predicate: WaitPredicateT = predicate or self.predicate
        self.__future: asyncio.Future[Any] = asyncio.Future()

    async def __call__(self, *args: Any) -> Any:
        if self.__future.done():
            self._set.discard(self)
            return

        try:
            result = await self._predicate(*args)
        except Exception as e:
            self._set.discard(self)
            self.__future.set_exception(e)
            return

        if not result:
            return

        self._set.discard(self)

        if not args:
            self.__future.set_result(None)
        else:
            self.__future.set_result(*args)

    async def wait(self) -> Any:
        try:
            async with asyncio.timeout(self._timeout):
                return await self.__future
        except TimeoutError:
            self._set.discard(self)
            raise

    async def predicate(self, *args: Any) -> bool:
        return True


F = TypeVar("F", bound=Callable[..., Any])


def handle_user_ids(is_self: bool = False) -> Callable[..., Any]:
    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            from .user import PartialUser

            new_args = [
                arg.id if isinstance(arg, PartialUser) else arg for i, arg in enumerate(args) if not (is_self and i == 0)
            ]

            if is_self and args:
                new_args.insert(0, args[0])

            new_kwargs = {k: (v.id if isinstance(v, PartialUser) else v) for k, v in kwargs.items()}

            return func(*new_args, **new_kwargs)

        return cast("F", wrapper)

    return decorator


def _is_submodule(parent: str, child: str) -> bool:
    return parent == child or child.startswith(f"{parent}.")


def date_to_datetime_with_z(date: datetime.date) -> str:
    return f"{datetime.datetime.combine(date, datetime.time(0, 0)).isoformat()}Z"


def flatten_literal_params(parameters: Iterable[Any]) -> tuple[Any, ...]:
    params: list[Any] = []
    literal_cls = type(Literal[0])

    for p in parameters:
        if isinstance(p, literal_cls):
            params.extend(p.__args__)  # type: ignore
        else:
            params.append(p)

    return tuple(params)


def normalise_optional_params(parameters: Iterable[Any]) -> tuple[Any, ...]:
    none_cls = type(None)
    return tuple(p for p in parameters if p is not none_cls) + (none_cls,)  # noqa: RUF005


def evaluate_annotation(
    tp: Any,
    globals: dict[str, Any],
    locals: dict[str, Any],
    cache: dict[str, Any],
    *,
    implicit_str: bool = True,
) -> Any:
    if isinstance(tp, ForwardRef):
        tp = tp.__forward_arg__
        # ForwardRefs always evaluate their internals
        implicit_str = True

    if implicit_str and isinstance(tp, str):
        if tp in cache:
            return cache[tp]
        evaluated = evaluate_annotation(eval(tp, globals, locals), globals, locals, cache)
        cache[tp] = evaluated
        return evaluated

    if PY_312 and getattr(tp.__repr__, "__objclass__", None) is typing.TypeAliasType:  # type: ignore
        temp_locals = dict(**locals, **{t.__name__: t for t in tp.__type_params__})  # type: ignore
        annotation = evaluate_annotation(tp.__value__, globals, temp_locals, cache.copy())  # type: ignore
        if hasattr(tp, "__args__"):
            annotation = annotation[tp.__args__]
        return annotation

    if hasattr(tp, "__supertype__"):
        return evaluate_annotation(tp.__supertype__, globals, locals, cache)

    if hasattr(tp, "__metadata__"):
        # Annotated[X, Y] can access Y via __metadata__
        metadata = tp.__metadata__[0]
        return evaluate_annotation(metadata, globals, locals, cache)

    if hasattr(tp, "__args__"):
        implicit_str = True
        is_literal = False
        args = tp.__args__
        if not hasattr(tp, "__origin__"):
            return tp
        if tp.__origin__ is Union:
            try:
                if args.index(type(None)) != len(args) - 1:
                    args = normalise_optional_params(tp.__args__)
            except ValueError:
                pass
        if tp.__origin__ is Literal:
            args = flatten_literal_params(tp.__args__)
            implicit_str = False
            is_literal = True

        evaluated_args = tuple(evaluate_annotation(arg, globals, locals, cache, implicit_str=implicit_str) for arg in args)

        if is_literal and not all(isinstance(x, (str, int, bool, type(None))) for x in evaluated_args):
            raise TypeError("Literal arguments must be of type str, int, bool, or NoneType.")

        try:
            return tp.copy_with(evaluated_args)
        except AttributeError:
            return tp.__origin__[evaluated_args]  # type: ignore

    return tp


def resolve_annotation(
    annotation: Any,
    globalns: dict[str, Any],
    localns: dict[str, Any] | None,
    cache: dict[str, Any] | None,
) -> Any:
    if annotation is None:
        return type(None)
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)

    locals = globalns if localns is None else localns
    if cache is None:
        cache = {}
    return evaluate_annotation(annotation, globalns, locals, cache)


def is_inside_class(func: Callable[..., Any]) -> bool:
    # For methods defined in a class, the qualname has a dotted path
    # denoting which class it belongs to. So, e.g. for A.foo the qualname
    # would be A.foo while a global foo() would just be foo.
    #
    # Unfortunately, for nested functions this breaks. So inside an outer
    # function named outer, those two would end up having a qualname with
    # outer.<locals>.A.foo and outer.<locals>.foo

    if func.__qualname__ == func.__name__:
        return False
    (remaining, _, _) = func.__qualname__.rpartition(".")
    return not remaining.endswith("<locals>")


def unwrap_function(function: Callable[..., Any], /) -> Callable[..., Any]:
    partial = functools.partial

    while True:
        if hasattr(function, "__wrapped__"):
            function = function.__wrapped__  # type: ignore
        elif isinstance(function, partial):
            function = function.func
        else:
            return function



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import (
        ChannelStreamScheduleResponseCategory,
        ChannelStreamScheduleResponseData,
        ChannelStreamScheduleResponseSegments,
        ChannelStreamScheduleResponseVacation,
    )


__all__ = ("Schedule", "ScheduleCategory", "ScheduleSegment", "ScheduleVacation")


class Schedule:
    """Represents a channel's stream schedule

    Attributes
    -----------
    segments: list[ScheduleSegment]
        The list of broadcasts in the broadcaster's streaming schedule.
    broadcaster: PartialUser
        The broadcaster that owns the broadcast schedule.
    vacation: ScheduleVacation | None
        The dates when the broadcaster is on vacation and not streaming. Is set to None if vacation mode is not enabled.
    """

    __slots__ = ("broadcaster", "segments", "vacation")

    def __init__(self, data: ChannelStreamScheduleResponseData, *, http: HTTPClient) -> None:
        self.segments: list[ScheduleSegment] = [
            ScheduleSegment(d, http=http, broadcaster_id=data["broadcaster_id"]) for d in data["segments"]
        ]
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.vacation: ScheduleVacation | None = ScheduleVacation(data["vacation"]) if data["vacation"] else None

    def __repr__(self) -> str:
        return f"<Schedule segments={self.segments} broadcaster={self.broadcaster} vacation={self.vacation}>"


class ScheduleSegment:
    """Represents a segment of a channel's stream schedule

    Attributes
    -----------
    id: str
        The ID for the scheduled broadcast.
    start_time: datetime.datetime
        Scheduled start time for the scheduled broadcast
    end_time: datetime.datetime
        Scheduled end time for the scheduled broadcast
    title: str
        Title for the scheduled broadcast.
    canceled_until: datetime.datetime | None
        Indicates whether the broadcaster canceled this segment of a recurring broadcast.
        If the broadcaster canceled this segment, this field is set to the same value that's in the end_time field; otherwise, it's None.
    category: ScheduleCategory | None
        The game or category details for the scheduled broadcast.
    is_recurring: bool
        Indicates if the scheduled broadcast is recurring weekly.
    """

    __slots__ = (
        "_broadcaster_id",
        "_http",
        "canceled_until",
        "category",
        "end_time",
        "id",
        "recurring",
        "start_time",
        "title",
    )

    def __init__(self, data: ChannelStreamScheduleResponseSegments, *, http: HTTPClient, broadcaster_id: str) -> None:
        self._http: HTTPClient = http
        self._broadcaster_id = broadcaster_id
        self.id: str = data["id"]
        self.start_time: datetime.datetime = parse_timestamp(data["start_time"])
        self.end_time: datetime.datetime = parse_timestamp(data["end_time"])
        self.title: str = data["title"]
        self.canceled_until: datetime.datetime | None = (
            parse_timestamp(data["canceled_until"]) if data["canceled_until"] else None
        )
        self.category: ScheduleCategory | None = ScheduleCategory(data["category"]) if data["category"] else None
        self.recurring: bool = bool(data["is_recurring"])

    def __repr__(self) -> str:
        return f"<ScheduleSegment id={self.id} start_time={self.start_time} end_time={self.end_time} title={self.title}>"

    async def update_segment(
        self,
        *,
        start_time: datetime.datetime | None = None,
        duration: int | None = None,
        category_id: str | None = None,
        title: str | None = None,
        canceled: bool | None = None,
        timezone: str | None = None,
    ) -> Schedule:
        """|coro|

        Updates a scheduled broadcast segment.

        .. note::
            Requires a user access token that includes the ``channel:manage:schedule`` scope.

        Parameters
        ----------
        token_for: str | PartialUser
            User access token that includes the ``channel:manage:schedule`` scope.
        start_time: datetime.datetime | None
            The datetime that the broadcast segment starts. This can be timezone aware.
        duration: int | None
            he length of time, in minutes, that the broadcast is scheduled to run. The duration must be in the range 30 through 1380 (23 hours)
        category_id: str | None
            The ID of the category that best represents the broadcast's content. To get the category ID, use the [Search Categories][twitchio.client.search_categories].
        title: str | None
            The broadcast's title. The title may contain a maximum of 140 characters.
        canceled: bool | None
            A Boolean value that indicates whether the broadcast is canceled. Set to True to cancel the segment.
        timezone: str | None
            The time zone where the broadcast takes place. Specify the time zone using `IANA time zone database <https://www.iana.org/time-zones>`_ format (for example, America/New_York).

        Returns
        -------
        Schedule
            Schedule object.

        Raises
        ------
        ValueError
            Duration must be between 30 and 1380.
        ValueError
            Title must not be greater than 140 characters.
        """
        if duration is not None and (duration < 30 or duration > 1380):
            raise ValueError("Duration must be between 30 and 1380.")
        if title is not None and len(title) > 140:
            raise ValueError("Title must not be greater than 140 characters.")

        data = await self._http.patch_channel_stream_schedule_segment(
            broadcaster_id=self._broadcaster_id,
            id=self.id,
            start_time=start_time,
            duration=duration,
            category_id=category_id,
            title=title,
            canceled=canceled,
            timezone=timezone,
            token_for=self._broadcaster_id,
        )

        return Schedule(data["data"], http=self._http)

    async def delete(self) -> None:
        """|coro|

        Removes a broadcast segment from the broadcaster's streaming schedule.

        For recurring segments, removing a segment removes all segments in the recurring schedule.

        .. note::
            Requires a user access token that includes the ``channel:manage:schedule`` scope.

        Parameters
        ----------
        token_for: str | PartialUser
            User access token that includes the ``channel:manage:schedule`` scope.
        """
        return await self._http.delete_channel_stream_schedule_segment(
            broadcaster_id=self._broadcaster_id, id=self.id, token_for=self._broadcaster_id
        )


class ScheduleCategory:
    """Game or category details of a stream's schedule

    Attributes
    -----------
    id: str
        The game or category ID.
    name: str
        The game or category name.
    """

    __slots__ = ("id", "name")

    def __init__(self, data: ChannelStreamScheduleResponseCategory) -> None:
        self.id = data["id"]
        self.name = data["name"]

    def __repr__(self) -> str:
        return f"<ScheduleCategory id={self.id} name={self.name}>"


class ScheduleVacation:
    """A schedule's vacation details

    Attributes
    -----------
    start_time: datetime.datetime
        Start date of stream schedule vaction.
    end_time: datetime.datetime
        End date of stream schedule vaction.
    """

    __slots__ = ("end_time", "start_time")

    def __init__(self, data: ChannelStreamScheduleResponseVacation) -> None:
        self.start_time = parse_timestamp(data["start_time"])
        self.end_time = parse_timestamp(data["end_time"])

    def __repr__(self) -> str:
        return f"<ScheduleVacation start_time={self.start_time} end_time={self.end_time}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from twitchio.assets import Asset
from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import ChannelTeamsResponseData, TeamsResponseData


__all__ = ("ChannelTeam", "Team")


class Team:
    """Represents information for a specific Twitch Team

    Attributes
    -----------
    users: list[PartialUser]
        List of users in the specified Team.
    background_image: Asset | None
        URL for the Team background image.
    banner: Asset | None
        URL for the Team banner.
    created_at: datetime.datetime
        Date and time the Team was created.
    updated_at: datetime.datetime
        Date and time the Team was last updated.
    info: str
        Team description.
    thumbnail: Asset | None
        Image URL for the Team logo.
    name: str
        Team name.
    display_name: str
        Team display name.
    id: str
        Team ID.
    """

    __slots__ = (
        "background_image",
        "banner",
        "created_at",
        "display_name",
        "id",
        "info",
        "name",
        "thumbnail",
        "updated_at",
        "users",
    )

    def __init__(self, data: TeamsResponseData, *, http: HTTPClient) -> None:
        self.users: list[PartialUser] = [
            PartialUser(x["user_id"], x["user_login"], x["user_name"], http=http) for x in data["users"]
        ]
        self.background_image: Asset | None = (
            Asset(data["background_image_url"], http=http) if data["background_image_url"] else None
        )
        self.banner: Asset | None = Asset(data["banner"], http=http) if data["banner"] else None
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.updated_at: datetime.datetime = parse_timestamp(data["updated_at"])
        self.info: str = data["info"]
        self.thumbnail: Asset | None = Asset(data["thumbnail_url"], http=http) if data["thumbnail_url"] else None
        self.name: str = data["team_name"]
        self.display_name: str = data["team_display_name"]
        self.id: str = data["id"]

    def __repr__(self) -> str:
        return f"<Team users={self.users} name={self.name} display_name={self.display_name} id={self.id} created_at={self.created_at}>"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, __value: object) -> bool:
        return __value.id == self.id if isinstance(__value, Team) else NotImplemented


class ChannelTeam:
    """Represents the Twitch Teams of which the specified channel/broadcaster is a member

    Attributes
    -----------
    broadcaster: PartialUser
        The broadcaster.
    background_image: Asset | None
        Asset for the team background image.
    banner: Asset | None
        Asset for the team banner.
    created_at: datetime.datetime
        Date and time the Team was created.
    updated_at: datetime.datetime`
        Date and time the Team was last updated.
    info: str
        Team description.
    thumbnail_url: Asset | None
        Asset for the team logo.
    name: str
        Team name.
    display_name: str
        Team display name.
    id: str
        Team ID.
    """

    __slots__ = (
        "background_image",
        "banner",
        "broadcaster",
        "created_at",
        "display_name",
        "id",
        "info",
        "name",
        "thumbnail",
        "updated_at",
    )

    def __init__(self, data: ChannelTeamsResponseData, http: HTTPClient) -> None:
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.background_image: Asset | None = (
            Asset(data["background_image_url"], http=http) if data["background_image_url"] else None
        )
        self.banner: Asset | None = Asset(data["banner"], http=http) if data["banner"] else None
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.updated_at: datetime.datetime = parse_timestamp(data["updated_at"])
        self.info: str = data["info"]
        self.thumbnail: Asset | None = Asset(data["thumbnail_url"], http=http) if data["thumbnail_url"] else None
        self.name: str = data["team_name"]
        self.display_name: str = data["team_display_name"]
        self.id: str = data["id"]

    def __repr__(self) -> str:
        return f"<ChannelTeam user={self.broadcaster} name={self.name} display_name={self.display_name} id={self.id} created_at={self.created_at}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


class _scope_property:
    def __set_name__(self, owner: type[Scopes], name: str) -> None:
        self._name = name

    def __get__(self, *_: object) -> _scope_property:
        return self

    def __set__(self, instance: Scopes, value: bool) -> None:
        if value is True:
            instance._selected.add(self)
        elif value is False:
            instance._selected.discard(self)
        else:
            raise TypeError(f"Expected bool for scope, got {type(value).__name__}")

    def __str__(self) -> str:
        return self._name.replace("_", ":", 2)

    def quoted(self) -> str:
        return urllib.parse.quote(str(self))

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self._name)

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, (_scope_property, str)):
            return NotImplemented

        return str(self) == str(other)


class _ScopeMeta(type):
    def __setattr__(self, name: str, value: object, /) -> None:
        raise AttributeError("Cannot set the value of a Scope property.")

    def __delattr__(self, name: str, /) -> None:
        raise AttributeError("Cannot delete the value of a Scope property.")


class Scopes(metaclass=_ScopeMeta):
    """The Scopes class is a helper utility class to help with selecting and formatting scopes for use in Twitch OAuth.

    The Scopes class can be initialised and used a few different ways:

    - Passing a ``list[str]`` of scopes to the constructor which can be in the same format as seen on Twitch. E.g. ``["user:read:email", ...]``.

    - Passing ``Keyword-Arguments`` to the constructor. E.g. ``user_read_email=True``

    - Or Constructing the class without passing anything and giving each required scope a bool when needed.

    - There is also a classmethod :meth:`.all` which selects all scopes for you.



    All scopes on this class are special descriptors.

    Attributes
    ----------
    analytics_read_extensions
        Equivalent to the ``analytics:read:extensions`` scope on Twitch.
    analytics_read_games
        Equivalent to the ``analytics:read:games`` scope on Twitch.
    bits_read
        Equivalent to the ``bits:read`` scope on Twitch.
    channel_bot
        Equivalent to the ``channel:bot`` scope on Twitch.
    channel_manage_ads
        Equivalent to the ``channel:manage:ads`` scope on Twitch.
    channel_read_ads
        Equivalent to the ``channel:read:ads`` scope on Twitch.
    channel_manage_broadcast
        Equivalent to the ``channel:manage:broadcast`` scope on Twitch.
    channel_read_charity
        Equivalent to the ``channel:read:charity`` scope on Twitch.
    channel_edit_commercial
        Equivalent to the ``channel:edit:commercial`` scope on Twitch.
    channel_read_editors
        Equivalent to the ``channel:read:editors`` scope on Twitch.
    channel_manage_extensions
        Equivalent to the ``channel:manage:extensions`` scope on Twitch.
    channel_read_goals
        Equivalent to the ``channel:read:goals`` scope on Twitch.
    channel_read_guest_star
        Equivalent to the ``channel:read:guest_star`` scope on Twitch.
    channel_manage_guest_star
        Equivalent to the ``channel:manage:guest_star`` scope on Twitch.
    channel_read_hype_train
        Equivalent to the ``channel:read:hype_train`` scope on Twitch.
    channel_manage_moderators
        Equivalent to the ``channel:manage:moderators`` scope on Twitch.
    channel_read_polls
        Equivalent to the ``channel:read:polls`` scope on Twitch.
    channel_manage_polls
        Equivalent to the ``channel:manage:polls`` scope on Twitch.
    channel_read_predictions
        Equivalent to the ``channel:read:predictions`` scope on Twitch.
    channel_manage_predictions
        Equivalent to the ``channel:manage:predictions`` scope on Twitch.
    channel_manage_raids
        Equivalent to the ``channel:manage:raids`` scope on Twitch.
    channel_read_redemptions
        Equivalent to the ``channel:read:redemptions`` scope on Twitch.
    channel_manage_redemptions
        Equivalent to the ``channel:manage:redemptions`` scope on Twitch.
    channel_manage_schedule
        Equivalent to the ``channel:manage:schedule`` scope on Twitch.
    channel_read_stream_key
        Equivalent to the ``channel:read:stream_key`` scope on Twitch.
    channel_read_subscriptions
        Equivalent to the ``channel:read:subscriptions`` scope on Twitch.
    channel_manage_videos
        Equivalent to the ``channel:manage:videos`` scope on Twitch.
    channel_read_vips
        Equivalent to the ``channel:read:vips`` scope on Twitch.
    channel_manage_vips
        Equivalent to the ``channel:manage:vips`` scope on Twitch.
    channel_moderate
        Equivalent to the ``channel:moderate`` scope on Twitch.
    clips_edit
        Equivalent to the ``clips:edit`` scope on Twitch.
    moderation_read
        Equivalent to the ``moderation:read`` scope on Twitch.
    moderator_manage_announcements
        Equivalent to the ``moderator:manage:announcements`` scope on Twitch.
    moderator_manage_automod
        Equivalent to the ``moderator:manage:automod`` scope on Twitch.
    moderator_read_automod_settings
        Equivalent to the ``moderator:read:automod_settings`` scope on Twitch.
    moderator_manage_automod_settings
        Equivalent to the ``moderator:manage:automod_settings`` scope on Twitch.
    moderator_read_banned_users
        Equivalent to the ``moderator:read:banned_users`` scope on Twitch.
    moderator_manage_banned_users
        Equivalent to the ``moderator:manage:banned_users`` scope on Twitch.
    moderator_read_blocked_terms
        Equivalent to the ``moderator:read:blocked_terms`` scope on Twitch.
    moderator_read_chat_messages
        Equivalent to the ``moderator:read:chat_messages`` scope on Twitch.
    moderator_manage_blocked_terms
        Equivalent to the ``moderator:manage:blocked_terms`` scope on Twitch.
    moderator_manage_chat_messages
        Equivalent to the ``moderator:manage:chat_messages`` scope on Twitch.
    moderator_read_chat_settings
        Equivalent to the ``moderator:read:chat_settings`` scope on Twitch.
    moderator_manage_chat_settings
        Equivalent to the ``moderator:manage:chat_settings`` scope on Twitch.
    moderator_read_chatters
        Equivalent to the ``moderator:read:chatters`` scope on Twitch.
    moderator_read_followers
        Equivalent to the ``moderator:read:followers`` scope on Twitch.
    moderator_read_guest_star
        Equivalent to the ``moderator:read:guest_star`` scope on Twitch.
    moderator_manage_guest_star
        Equivalent to the ``moderator:manage:guest_star`` scope on Twitch.
    moderator_read_moderators
        Equivalent to the ``moderator:read:moderators`` scope on Twitch.
    moderator_read_shield_mode
        Equivalent to the ``moderator:read:shield_mode`` scope on Twitch.
    moderator_manage_shield_mode
        Equivalent to the ``moderator:manage:shield_mode`` scope on Twitch.
    moderator_read_shoutouts
        Equivalent to the ``moderator:read:shoutouts`` scope on Twitch.
    moderator_manage_shoutouts
        Equivalent to the ``moderator:manage:shoutouts`` scope on Twitch.
    moderator_read_suspicious_users
        Equivalent to the ``moderator:read:suspicious_users`` scope on Twitch.
    moderator_read_unban_requests
        Equivalent to the ``moderator:read:unban_requests`` scope on Twitch.
    moderator_manage_unban_requests
        Equivalent to the ``moderator:manage:unban_requests`` scope on Twitch.
    moderator_read_vips
        Equivalent to the ``moderator:read:vips`` scope on Twitch.
    moderator_read_warnings
        Equivalent to the ``moderator:read:warnings`` scope on Twitch.
    moderator_manage_warnings
        Equivalent to the ``moderator:manage:warnings`` scope on Twitch.
    user_bot
        Equivalent to the ``user:bot`` scope on Twitch.
    user_edit
        Equivalent to the ``user:edit`` scope on Twitch.
    user_edit_broadcast
        Equivalent to the ``user:edit:broadcast`` scope on Twitch.
    user_read_blocked_users
        Equivalent to the ``user:read:blocked_users`` scope on Twitch.
    user_manage_blocked_users
        Equivalent to the ``user:manage:blocked_users`` scope on Twitch.
    user_read_broadcast
        Equivalent to the ``user:read:broadcast`` scope on Twitch.
    user_read_chat
        Equivalent to the ``user:read:chat`` scope on Twitch.
    user_manage_chat_color
        Equivalent to the ``user:manage:chat_color`` scope on Twitch.
    user_read_email
        Equivalent to the ``user:read:email`` scope on Twitch.
    user_read_emotes
        Equivalent to the ``user:read:emotes`` scope on Twitch.
    user_read_follows
        Equivalent to the ``user:read:follows`` scope on Twitch.
    user_read_moderated_channels
        Equivalent to the ``user:read:moderated_channels`` scope on Twitch.
    user_read_subscriptions
        Equivalent to the ``user:read:subscriptions`` scope on Twitch.
    user_read_whispers
        Equivalent to the ``user:read:whispers`` scope on Twitch.
    user_manage_whispers
        Equivalent to the ``user:manage:whispers`` scope on Twitch.
    user_write_chat
        Equivalent to the ``user:write:chat`` scope on Twitch.
    chat_edit
        Equivalent to the ``chat:edit`` scope on Twitch.
    chat_read
        Equivalent to the ``chat:read`` scope on Twitch.
    user_edit_follows
        Equivalent to the ``user:edit:follows`` scope on Twitch.
    whispers_read
        Equivalent to the ``whispers:read`` scope on Twitch.
    whispers_edit
        Equivalent to the ``whispers:edit`` scope on Twitch.
    """

    __slots__ = ("_selected",)

    analytics_read_extensions = _scope_property()
    analytics_read_games = _scope_property()
    bits_read = _scope_property()
    channel_bot = _scope_property()
    channel_manage_ads = _scope_property()
    channel_read_ads = _scope_property()
    channel_manage_broadcast = _scope_property()
    channel_read_charity = _scope_property()
    channel_edit_commercial = _scope_property()
    channel_read_editors = _scope_property()
    channel_manage_extensions = _scope_property()
    channel_read_goals = _scope_property()
    channel_read_guest_star = _scope_property()
    channel_manage_guest_star = _scope_property()
    channel_read_hype_train = _scope_property()
    channel_manage_moderators = _scope_property()
    channel_read_polls = _scope_property()
    channel_manage_polls = _scope_property()
    channel_read_predictions = _scope_property()
    channel_manage_predictions = _scope_property()
    channel_manage_raids = _scope_property()
    channel_read_redemptions = _scope_property()
    channel_manage_redemptions = _scope_property()
    channel_manage_schedule = _scope_property()
    channel_read_stream_key = _scope_property()
    channel_read_subscriptions = _scope_property()
    channel_manage_videos = _scope_property()
    channel_read_vips = _scope_property()
    channel_manage_vips = _scope_property()
    channel_moderate = _scope_property()
    clips_edit = _scope_property()
    moderation_read = _scope_property()
    moderator_manage_announcements = _scope_property()
    moderator_manage_automod = _scope_property()
    moderator_read_automod_settings = _scope_property()
    moderator_manage_automod_settings = _scope_property()
    moderator_read_banned_users = _scope_property()
    moderator_manage_banned_users = _scope_property()
    moderator_read_blocked_terms = _scope_property()
    moderator_read_chat_messages = _scope_property()
    moderator_manage_blocked_terms = _scope_property()
    moderator_manage_chat_messages = _scope_property()
    moderator_read_chat_settings = _scope_property()
    moderator_manage_chat_settings = _scope_property()
    moderator_read_chatters = _scope_property()
    moderator_read_followers = _scope_property()
    moderator_read_guest_star = _scope_property()
    moderator_manage_guest_star = _scope_property()
    moderator_read_moderators = _scope_property()
    moderator_read_shield_mode = _scope_property()
    moderator_manage_shield_mode = _scope_property()
    moderator_read_shoutouts = _scope_property()
    moderator_manage_shoutouts = _scope_property()
    moderator_read_suspicious_users = _scope_property()
    moderator_read_unban_requests = _scope_property()
    moderator_manage_unban_requests = _scope_property()
    moderator_read_vips = _scope_property()
    moderator_read_warnings = _scope_property()
    moderator_manage_warnings = _scope_property()
    user_bot = _scope_property()
    user_edit = _scope_property()
    user_edit_broadcast = _scope_property()
    user_read_blocked_users = _scope_property()
    user_manage_blocked_users = _scope_property()
    user_read_broadcast = _scope_property()
    user_read_chat = _scope_property()
    user_manage_chat_color = _scope_property()
    user_read_email = _scope_property()
    user_read_emotes = _scope_property()
    user_read_follows = _scope_property()
    user_read_moderated_channels = _scope_property()
    user_read_subscriptions = _scope_property()
    user_read_whispers = _scope_property()
    user_manage_whispers = _scope_property()
    user_write_chat = _scope_property()
    chat_edit = _scope_property()
    chat_read = _scope_property()
    user_edit_follows = _scope_property()
    whispers_read = _scope_property()
    whispers_edit = _scope_property()

    def __init__(self, scopes: Iterable[str | _scope_property] | None = None, /, **kwargs: bool) -> None:
        if scopes is None:
            scopes = []
        self._selected: set[_scope_property] = set()

        prop: _scope_property

        for scope in scopes:
            if isinstance(scope, str):
                prop = getattr(self, scope.replace(":", "_"))
            elif isinstance(scope, _scope_property):  # type: ignore
                prop = scope
            else:
                raise TypeError(f"Invalid scope provided: {type(scope)} is not a valid scope.")

            self._selected.add(prop)

        for key, value in kwargs.items():
            prop = getattr(self, key)

            if value is True:
                self._selected.add(prop)
            elif value is False:
                self._selected.discard(prop)
            else:
                raise TypeError(f'Expected bool for scope kwarg "{key}", got {type(value).__name__}')

    def __iter__(self) -> Iterator[str]:
        return iter([str(scope) for scope in self._selected])

    def __repr__(self) -> str:
        return f"<Scopes selected={list(self)}>"

    def __str__(self) -> str:
        return self.urlsafe()

    def __contains__(self, scope: _scope_property | str, /) -> bool:
        if isinstance(scope, str):
            return any(s.value == scope for s in self._selected)

        return scope in self._selected

    def urlsafe(self, *, unquote: bool = False) -> str:
        """Method which returns a URL-Safe formatted ``str`` of selected scopes.

        The string returned by this method is safe to use in browsers etc.

        Parameters
        ----------
        unqoute: bool
            If this is ``True``, this will return scopes without URL quoting, E.g. as ``user:read:email+channel:bot``
            compared to ``user%3Aread%3Aemail+channel%3Abot``. Defaults to ``False``.
        """
        return "+".join([scope.value if unquote else scope.quoted() for scope in self._selected])

    @property
    def selected(self) -> list[str]:
        """Property that returns a ``list[str]`` of selected scopes.

        This is not URL-Safe. See: :meth:`.urlsafe` for a method which returns a URL-Safe string.
        """
        return list(self)

    @classmethod
    def all(cls) -> Scopes:
        """Classmethod which creates this :class:`.Scopes` object with all scopes selected."""
        return cls([scope for scope in cls.__dict__.values() if isinstance(scope, _scope_property)])



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import secrets
import urllib.parse
from typing import TYPE_CHECKING, ClassVar

from ..http import HTTPClient, Route
from ..utils import MISSING
from .payloads import *


if TYPE_CHECKING:
    import aiohttp

    from ..types_.responses import (
        AuthorizationURLResponse,
        ClientCredentialsResponse,
        RefreshTokenResponse,
        UserTokenResponse,
        ValidateTokenResponse,
    )
    from .scopes import Scopes


class OAuth(HTTPClient):
    CONTENT_TYPE_HEADER: ClassVar[dict[str, str]] = {"Content-Type": "application/x-www-form-urlencoded"}

    def __init__(
        self,
        *,
        client_id: str,
        client_secret: str,
        redirect_uri: str | None = None,
        scopes: Scopes | None = None,
        session: aiohttp.ClientSession = MISSING,
    ) -> None:
        super().__init__(session=session, client_id=client_id)

        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scopes = scopes

    async def validate_token(self, token: str, /) -> ValidateTokenPayload:
        token = token.removeprefix("Bearer ").removeprefix("OAuth ")

        headers: dict[str, str] = {"Authorization": f"OAuth {token}"}
        route: Route = Route("GET", "/oauth2/validate", use_id=True, headers=headers)

        data: ValidateTokenResponse = await self.request_json(route)
        return ValidateTokenPayload(data)

    async def refresh_token(self, refresh_token: str, /) -> RefreshTokenPayload:
        params = self._create_params(
            {
                "grant_type": "refresh_token",
                "refresh_token": urllib.parse.quote(refresh_token, safe=""),
            }
        )

        route: Route = Route("POST", "/oauth2/token", use_id=True, headers=self.CONTENT_TYPE_HEADER, params=params)
        data: RefreshTokenResponse = await self.request_json(route)

        return RefreshTokenPayload(data)

    async def user_access_token(self, code: str, /, *, redirect_uri: str | None = None) -> UserTokenPayload:
        redirect = redirect_uri or self.redirect_uri
        if not redirect:
            raise ValueError('"redirect_uri" is a required parameter or attribute which is missing.')

        params = self._create_params(
            {
                "code": code,
                "grant_type": "authorization_code",
                "redirect_uri": redirect,
                # "scope": " ".join(SCOPES), #TODO
                # "state": #TODO
            }
        )

        route: Route = Route("POST", "/oauth2/token", use_id=True, headers=self.CONTENT_TYPE_HEADER, params=params)
        data: UserTokenResponse = await self.request_json(route)

        return UserTokenPayload(data)

    async def revoke_token(self, token: str, /) -> None:
        params = self._create_params({"token": token})

        route: Route = Route("POST", "/oauth2/revoke", use_id=True, headers=self.CONTENT_TYPE_HEADER, params=params)
        await self.request_json(route)

    async def client_credentials_token(self) -> ClientCredentialsPayload:
        params = self._create_params({"grant_type": "client_credentials"})

        route: Route = Route("POST", "/oauth2/token", use_id=True, headers=self.CONTENT_TYPE_HEADER, params=params)
        data: ClientCredentialsResponse = await self.request_json(route)

        return ClientCredentialsPayload(data)

    def get_authorization_url(
        self,
        *,
        scopes: Scopes | None = None,
        state: str | None = None,
        redirect_uri: str | None = None,
        force_verify: bool = False,
    ) -> AuthorizationURLPayload:
        redirect = redirect_uri or self.redirect_uri
        if not redirect:
            raise ValueError('"redirect_uri" is a required parameter or attribute which is missing.')

        scopes = scopes or self.scopes
        if not scopes:
            raise ValueError('"scopes" is a required parameter or attribute which is missing.')

        if state is None:
            state = secrets.token_urlsafe(32)

        params = {
            "client_id": self.client_id,
            "redirect_uri": urllib.parse.quote(redirect),
            "response_type": "code",
            "scope": scopes.urlsafe(),
            "force_verify": "true" if force_verify else "false",
            "state": state,
        }

        route: Route = Route("GET", "/oauth2/authorize", use_id=True, params=params)
        data: AuthorizationURLResponse = {
            "url": route.url,
            "client_id": self.client_id,
            "redirect_uri": redirect,
            "response_type": "code",
            "scopes": scopes.selected,
            "force_verify": force_verify,
            "state": state,
        }

        payload: AuthorizationURLPayload = AuthorizationURLPayload(data)
        return payload

    def _create_params(self, extra_params: dict[str, str]) -> dict[str, str]:
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }
        params.update(extra_params)
        return params



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import datetime
import logging
from collections import deque
from typing import TYPE_CHECKING, Any, cast
from urllib.parse import unquote_plus

import uvicorn
from starlette.applications import Starlette
from starlette.responses import RedirectResponse, Response
from starlette.routing import Route

from ..authentication import Scopes
from ..eventsub.subscriptions import _SUB_MAPPING
from ..exceptions import HTTPException
from ..models.eventsub_ import SubscriptionRevoked, create_event_instance
from ..utils import _from_json, parse_timestamp  # type: ignore
from .utils import MESSAGE_TYPES, BaseAdapter, FetchTokenPayload, verify_message


if TYPE_CHECKING:
    from starlette.requests import Request

    from ..authentication import AuthorizationURLPayload, UserTokenPayload
    from ..client import Client
    from ..types_.eventsub import EventSubHeaders


__all__ = ("StarletteAdapter",)


logger: logging.Logger = logging.getLogger(__name__)


class StarletteAdapter(BaseAdapter, Starlette):
    """The StarletteAdapter for OAuth and Webhook based EventSub.

    This adapter uses ``starlette`` which is an optional dependency and needs to be installed.

    Optionally you can use `Aiohttp <https://docs.aiohttp.org/en/stable/>`_  by using the :class:`AiohttpAdapter`.

    An adapter will always be started and is ran alongside the :class:`~twitchio.Client` or
    :class:`~twitchio.ext.commands.Bot` by default. You can disable starting the adapter by passing ``with_adapter=False``
    to :meth:`twitchio.Client.start` or :meth:`twitchio.Client.run`.

    The adapter can be used to authenticate users via OAuth, and supports webhooks for EventSub, with in-built event
    dispatching to the :class:`~twitchio.Client`.

    The default callbacks for OAuth are:

    - ``/oauth`` E.g. ``http://localhost:4343/oauth`` or ``https://mydomain.org/oauth``.
        - Should be used with the ``?scopes=`` query parameter to pass a URL encoded list of scopes.
           See: :class:`~twitchio.Scopes` for a helper class for generating scopes.

    - ``/oauth/callback`` E.g. ``http://localhost:4343/oauth/callback`` or ``https://mydomain.org/oauth/callback``.
        - The redirect URL for OAuth request. This should be set in the developer console of your application on Twitch.

    The default callbacks for EventSub are:

    - ``/callback`` E.g. ``http://localhost:4343/callback`` or ``https://mydomain.org/callback``.


    This class handles processing and validating EventSub requests for you, and dispatches any events identical to
    the websocket equivalent.

    Parameters
    ----------
    host: str | None
        An optional :class:`str` passed to bind as the host address. Defaults to ``localhost``.
    port: int | None
        An optional :class:`int` passed to bind as the port for the host address. Defaults to ``4343``.
    domain: str | None
        An optional :class:`str` passed used to identify the external domain used, if any. If passed, the domain will be used
        for the redirect URL in OAuth and for validation in EventSub. It must be publicly accessible and support HTTPS.
    eventsub_path: str | None
        An optional :class:`str` passed to use as the path to the eventsub callback. Defaults to ``/callback``. E.g.
        ``http://localhost:4343/callback`` or ``https://mydomain.org/callback``.
    eventsub_secret: str | None
        An optional :class:`str` passed to use as the EventSub secret. It is recommended you pass this parameter when using
        an adapter for EventSub, as it will reset upon restarting otherwise. You can generate token safe secrets with the
        :mod:`secrets` module.

    Examples
    --------

    .. code-block:: python3

        import twitchio
        from twitchio import web
        from twitchio.ext import commands


        class Bot(commands.Bot):

            def __init__(self) -> None:
                # Requests will be sent to, as an example:
                # http://bot.twitchio.dev
                # You DO NOT need to pass a domain, however not passing a domain will mean only you can authenticate
                # via OAuth and Webhooks will not be supported...
                # The domain should support HTTPS and be publicly accessible...
                # An easy way to do this is add your domain to a CDN like Cloudflare and set SSL to Flexible.
                #
                # Visit: http://localhost:8080/oauth?scopes=channel:bot as the broadcaster as an example.
                # Visit: https://bot.twitchio.dev?scopes=channel:bot as the broadcaster as an example.

                adapter = web.StarletteAdapter(domain="bot.twitchio.dev", port=8080)
                super().__init__(adapter=adapter)
    """

    client: Client

    def __init__(
        self,
        *,
        host: str | None = None,
        port: int | None = None,
        domain: str | None = None,
        eventsub_path: str | None = None,
        eventsub_secret: str | None = None,
    ) -> None:
        self._host: str = host or "localhost"
        self._port: int = port or 4343

        self._eventsub_secret: str | None = eventsub_secret
        if eventsub_secret and not 10 <= len(eventsub_secret) <= 100:
            raise ValueError("Eventsub Secret must be between 10 and 100 characters long.")

        self._domain: str | None = None
        if domain:
            domain_ = domain.removeprefix("http://").removeprefix("https://").removesuffix("/")
            self._domain = f"https://{domain_}"
        else:
            self._domain = f"http://{self._host}:{self._port}"

        path: str = eventsub_path.removeprefix("/").removesuffix("/") if eventsub_path else "callback"
        self._eventsub_path: str = f"/{path}"

        self._runner_task: asyncio.Task[None] | None = None
        self._responded: deque[str] = deque(maxlen=5000)

        super().__init__(
            routes=[
                Route("/oauth/callback", self.oauth_callback, methods=["GET"]),
                Route("/oauth", self.oauth_redirect, methods=["GET"]),
                Route(self._eventsub_path, self.eventsub_callback, methods=["POST"]),
            ],
            on_shutdown=[self.event_shutdown],
            on_startup=[self.event_startup],
        )

    def __repr__(self) -> str:
        return f"StarletteAdapter(host={self._host}, port={self._port})"

    @property
    def eventsub_url(self) -> str:
        """Property returning the fully qualified URL to the EventSub callback."""
        return f"{self._domain}{self._eventsub_path}"

    @property
    def redirect_url(self) -> str:
        """Property returning the fully qualified URL to the OAuth callback."""
        return f"{self._domain}/oauth/callback"

    async def event_startup(self) -> None:
        logger.info("Starting TwitchIO StarletteAdapter on http://%s:%s.", self._host, self._port)

    async def event_shutdown(self) -> None:
        await self.close()

    async def close(self) -> None:
        if self._runner_task is not None:
            try:
                self._runner_task.cancel()
            except Exception as e:
                logger.debug(
                    "Ignoring exception raised while cancelling runner in <%s>: %s.",
                    self.__class__.__qualname__,
                    e,
                )

            self._runner_task = None
            await self.client.close()

        logger.info("Successfully shutdown TwitchIO <%s>.", self.__class__.__qualname__)

    async def run(self, host: str | None = None, port: int | None = None) -> None:
        self._host = host or self._host
        self._port = port or self._port

        config: uvicorn.Config = uvicorn.Config(
            app=self,
            host=self._host,
            port=self._port,
            log_level="critical",
            workers=0,
            timeout_graceful_shutdown=3,
        )

        server: uvicorn.Server = uvicorn.Server(config)
        self._runner_task = asyncio.create_task(server.serve(), name=f"twitchio-web-adapter:{self.__class__.__qualname__}")

    async def eventsub_callback(self, request: Request) -> Response:
        headers: EventSubHeaders = cast("EventSubHeaders", request.headers)
        msg_type: str | None = headers.get("Twitch-Eventsub-Message-Type")

        if not msg_type or msg_type not in MESSAGE_TYPES:
            logger.debug("Eventsub Webhook received an unknown Message-Type header value.")
            return Response(status_code=400)

        if not self._eventsub_secret:
            msg: str = f"Eventsub Webhook '{self!r}' must be passed a secret. See: ... for more info.'"
            return Response(msg, status_code=400)

        msg_id: str | None = headers.get("Twitch-Eventsub-Message-Id", None)
        timestamp: str | None = headers.get("Twitch-Eventsub-Message-Timestamp", None)

        if not msg_id or not timestamp:
            return Response("Bad Request. Invalid Message-ID or Message-Timestamp.", status_code=400)

        if msg_id in self._responded:
            return Response("Previously responded to Message.", status_code=400)

        self._responded.append(msg_id)

        try:
            resp: bytes = await verify_message(request=request, secret=self._eventsub_secret)
        except ValueError:
            return Response("Challenge Failed. Failed to verify the integrity of the message.", status_code=400)
        except Exception as e:
            return Response(f"Challenge Failed. Failed to verify the integrity of the message: {e}", status_code=400)

        data: Any = _from_json(resp)  # type: ignore
        sent: datetime.datetime = parse_timestamp(timestamp)
        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)

        if sent + datetime.timedelta(minutes=10) <= now:
            return Response("Message has expired.", status_code=400)

        if msg_type == "webhook_callback_verification":
            return Response(data["challenge"], status_code=200, headers={"Content-Type": "text/plain"})

        elif msg_type == "notification":
            sub_type: str = data["subscription"]["type"]
            event = _SUB_MAPPING.get(sub_type, sub_type.removeprefix("channel.")).replace(".", "_")

            try:
                payload_class = create_event_instance(sub_type, data, http=self.client._http, headers=headers)
            except ValueError:
                logger.warning("Webhook '%s' received an unhandled eventsub event: '%s'.", self, event)
                return Response(status_code=200)

            self.client.dispatch(event=event, payload=payload_class)
            return Response(status_code=200)

        elif msg_type == "revocation":
            payload: SubscriptionRevoked = SubscriptionRevoked(data["subscription"])
            self.client.dispatch(event="subscription_revoked", payload=payload)

            return Response(status_code=204)

    async def fetch_token(self, request: Request) -> FetchTokenPayload:
        """This method handles sending the provided code to Twitch to receive a User Access and Refresh Token pair, and
        later, if successful, dispatches :func:`~twitchio.event_oauth_authorized`.

        To call this coroutine you should pass the request received in the :meth:`oauth_callback`. This method is called by
        default, however when overriding :meth:`oauth_callback` you should always call this method.

        Parameters
        ----------
        request: starlette.requests.Request
            The request received in :meth:`oauth_callback`.

        Returns
        -------
        FetchTokenPayload
            The payload containing various information about the authentication request to Twitch.
        """
        if "code" not in request.query_params:
            return FetchTokenPayload(400, response=Response(status_code=400, content="No 'code' parameter provided."))

        try:
            resp: UserTokenPayload = await self.client._http.user_access_token(
                request.query_params["code"],
                redirect_uri=self.redirect_url,
            )
        except HTTPException as e:
            logger.error("Exception raised while fetching Token in <%s>: %s", self.__class__.__qualname__, e)
            status: int = e.status
            return FetchTokenPayload(status=status, response=Response(status_code=status), exception=e)

        self.client.dispatch(event="oauth_authorized", payload=resp)
        return FetchTokenPayload(
            status=20,
            response=Response(content="Success. You can leave this page.", status_code=200),
            payload=resp,
        )

    async def oauth_callback(self, request: Request) -> Response:
        """Default route callback for the OAuth Authentication redirect URL.

        You can override this method to alter the responses sent to the user.

        This callback should always return a valid response. See: `Starlette Responses <https://www.starlette.io/responses/>`_
        for available response types.

        .. important::

            You should always call :meth:`.fetch_token` when overriding this method.

        Parameters
        ----------
        request: starlette.requests.Request
            The original request received via Starlette.

        Examples
        --------

        .. code:: python3

            async def oauth_callback(self, request: Request) -> Response:
                payload: FetchTokenPayload = await self.fetch_token(request)

                # Change the default success response...
                if payload.status == 200:
                    return HTMLResponse(status_code=200, "<h1>Success!</h1>")

                # Return the default error responses...
                return payload.response
        """
        logger.debug("Received OAuth callback request in <%s>.", self.oauth_callback.__qualname__)

        payload: FetchTokenPayload = await self.fetch_token(request)
        if not isinstance(payload.response, Response):
            raise ValueError(f"Responses in StarlettepAdapter should be {type(Response)!r} not {type(payload.response)!r}")

        return payload.response

    async def oauth_redirect(self, request: Request) -> Response:
        scopes: str | None = request.query_params.get("scopes", None)
        force_verify: bool = request.query_params.get("force_verify", "false").lower() == "true"

        if not scopes:
            scopes = str(self.client._http.scopes) if self.client._http.scopes else None

        if not scopes:
            logger.warning(
                "No scopes provided in request to <%s>. Scopes are a required parameter that is missing.",
                self.oauth_redirect.__qualname__,
            )
            return Response("No scopes were provided. Scopes must be provided.", status_code=400)

        scopes_: Scopes = Scopes(unquote_plus(scopes).split())

        try:
            payload: AuthorizationURLPayload = self.client._http.get_authorization_url(
                scopes=scopes_,
                redirect_uri=self.redirect_url,
                force_verify=force_verify,
            )
        except Exception as e:
            logger.error("Exception raised while fetching Authorization URL in <%s>: %s", self.__class__.__qualname__, e)
            return Response(status_code=500)

        return RedirectResponse(url=payload["url"], status_code=307)



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
from typing import Any

from twitchio.exceptions import TwitchioException

from .cooldowns import BaseCooldown


__all__ = (
    "ArgumentError",
    "BadArgument",
    "CommandError",
    "CommandExistsError",
    "CommandHookError",
    "CommandInvokeError",
    "CommandNotFound",
    "CommandOnCooldown",
    "ComponentLoadError",
    "ConversionError",
    "ExpectedClosingQuoteError",
    "GuardFailure",
    "InputError",
    "InvalidEndOfQuotedStringError",
    "MissingRequiredArgument",
    "ModuleAlreadyLoadedError",
    "ModuleLoadFailure",
    "ModuleNotLoadedError",
    "NoEntryPointError",
    "PrefixError",
    "UnexpectedQuoteError",
)


class CommandError(TwitchioException):
    """Base exception for command related errors.

    All commands.ext related exceptions inherit from this class.
    """


class ComponentLoadError(CommandError):
    """Exception raised when a :class:`.commands.Component` fails to load."""


class CommandInvokeError(CommandError):
    """Exception raised when an error occurs during invocation of a command.

    Attributes
    ----------
    original: :class:`Exception` | None
        The original exception that caused this error. Could be None.
    """

    def __init__(self, msg: str | None = None, original: Exception | None = None) -> None:
        self.original: Exception | None = original
        super().__init__(msg)


class CommandHookError(CommandInvokeError): ...


class CommandNotFound(CommandError):
    """Exception raised when a message is processed with a valid prefix and no :class:`~twitchio.ext.commands.Command`
    could be found.
    """


class CommandExistsError(CommandError):
    """Exception raised when you try to add a command or alias to a command that is already registered on the
    :class:`~twitchio.ext.commands.Bot`."""


class PrefixError(CommandError):
    """Exception raised when invalid prefix or prefix callable is passed."""


class InputError(CommandError):
    """Base exception for errors raised while parsing the input for command invocation. All :class:`ArgumentError` and
    child exception inherit from this class."""


class ArgumentError(InputError):
    """Base exception for errors raised while parsing arguments in commands."""


class UnexpectedQuoteError(ArgumentError):
    def __init__(self, quote: str) -> None:
        self.quote: str = quote
        super().__init__(f"Unexpected quote mark, {quote!r}, in non-quoted string")


class InvalidEndOfQuotedStringError(ArgumentError):
    def __init__(self, char: str) -> None:
        self.char: str = char
        super().__init__(f"Expected space after closing quotation but received {char!r}")


class ExpectedClosingQuoteError(ArgumentError):
    def __init__(self, close_quote: str) -> None:
        self.close_quote: str = close_quote
        super().__init__(f"Expected closing {close_quote}.")


class GuardFailure(CommandError):
    """Exception raised when a :func:`~.commands.guard` fails or blocks a command from executing.

    This exception should be subclassed when raising a custom exception for a :func:`~twitchio.ext.commands.guard`.
    """

    def __init__(self, msg: str | None = None, *, guard: Any | None = None) -> None:
        self.guard: Any | None = guard
        super().__init__(msg or "")


class ConversionError(ArgumentError):
    """Base exception for conversion errors which occur during argument parsing in commands."""


class BadArgument(ConversionError):
    """Exception raised when a parsing or conversion failure is encountered on an argument to pass into a command."""

    def __init__(self, msg: str, *, name: str | None = None, value: str | None) -> None:
        self.name: str | None = name
        self.value: str | None = value
        super().__init__(msg)


class MissingRequiredArgument(ArgumentError):
    """Exception raised when parsing a command and a parameter that is required is not encountered."""

    def __init__(self, param: inspect.Parameter) -> None:
        self.param: inspect.Parameter = param
        super().__init__(f'"{param.name}" is a required argument which is missing.')


class ModuleError(TwitchioException):
    """Base exception for module related errors."""


class ModuleLoadFailure(ModuleError):
    """Exception raised when a module failed to load during execution or `setup` entry point."""

    def __init__(self, name: str, exc: Exception) -> None:
        super().__init__(name, exc)


class NoEntryPointError(ModuleError):
    """Exception raised when the module does not have a `setup` entry point coroutine."""

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class ModuleAlreadyLoadedError(ModuleError):
    """Exception raised when a module has already been loaded."""

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class ModuleNotLoadedError(ModuleError):
    """Exception raised when a module was not loaded."""

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class CommandOnCooldown(GuardFailure):
    """Exception raised when a command is invoked while on cooldown/ratelimited.

    Attributes
    ----------
    cooldown: :class:`~.commands.BaseCooldown`
        The specific cooldown instance used that raised this error.
    remaining: :class:`float`
        The time remaining for the cooldown as a :class:`float` of seconds.
    """

    def __init__(self, msg: str | None = None, *, cooldown: BaseCooldown, remaining: float) -> None:
        self.cooldown: BaseCooldown = cooldown
        self.remaining: float = remaining
        super().__init__(msg or f"Cooldown is ratelimited. Try again in {remaining} seconds.")



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import datetime
import logging
from typing import TYPE_CHECKING, cast

import aiohttp

from ..backoff import Backoff
from ..exceptions import HTTPException, WebsocketConnectionException
from ..models.eventsub_ import SubscriptionRevoked, create_event_instance
from ..utils import (
    MISSING,
    _from_json,  # type: ignore
)
from .subscriptions import _SUB_MAPPING


if TYPE_CHECKING:
    from ..authentication.tokens import ManagedHTTPClient
    from ..client import Client
    from ..types_.conduits import (
        Condition,
        MessageTypes,
        MetaData,
        NotificationMessage,
        ReconnectMessage,
        RevocationMessage,
        WebsocketMessages,
        WelcomeMessage,
        WelcomePayload,
    )
    from ..types_.eventsub import SubscriptionResponse, _SubscriptionData


logger: logging.Logger = logging.getLogger(__name__)


WSS: str = "wss://eventsub.wss.twitch.tv/ws"


class Websocket:
    __slots__ = (
        "__subscription_cost",
        "_backoff",
        "_client",
        "_closed",
        "_closing",
        "_connecting",
        "_connection_tasks",
        "_heartbeat",
        "_http",
        "_keep_alive_task",
        "_keep_alive_timeout",
        "_last_keepalive",
        "_listen_task",
        "_original_attempts",
        "_ready",
        "_reconnect_attempts",
        "_session_id",
        "_socket",
        "_subscriptions",
        "_token_for",
    )

    def __init__(
        self,
        *,
        keep_alive_timeout: float = 10,
        reconnect_attempts: int | None = MISSING,
        client: Client | None = None,
        token_for: str,
        http: ManagedHTTPClient,
    ) -> None:
        self._keep_alive_timeout: int = max(10, min(int(keep_alive_timeout), 600))
        self._heartbeat: int = min(self._keep_alive_timeout, 25) + 5
        self._last_keepalive: datetime.datetime | None = None
        self._keep_alive_task: asyncio.Task[None] | None = None

        self._session_id: str | None = None

        self._socket: aiohttp.ClientWebSocketResponse | None = None
        self._listen_task: asyncio.Task[None] | None = None

        self._ready: asyncio.Event = asyncio.Event()

        attempts: int | None = (
            0 if reconnect_attempts is None else None if reconnect_attempts is MISSING else reconnect_attempts
        )
        self._original_attempts = reconnect_attempts
        self._reconnect_attempts = attempts
        self._backoff: Backoff = Backoff(base=3, maximum_time=90)

        self.__subscription_cost: int = 0

        self._client: Client | None = client
        self._token_for: str = token_for
        self._http: ManagedHTTPClient = http
        self._subscriptions: dict[str, _SubscriptionData] = {}

        self._connecting: bool = False
        self._closed: bool = False
        self._closing: bool = False

        self._connection_tasks: set[asyncio.Task[None]] = set()

        msg = "Websocket %s is being used without a Client/Bot. Event dispatching is disabled for this websocket."
        if not client:
            logger.warning(msg, self)

    def __repr__(self) -> str:
        return f"EventsubWebsocket(session_id={self._session_id})"

    def __str__(self) -> str:
        return f"{self._session_id}"

    @property
    def keep_alive_timeout(self) -> int:
        return self._keep_alive_timeout

    @property
    def connected(self) -> bool:
        return bool(self._socket and not self._socket.closed)

    @property
    def session_id(self) -> str | None:
        return self._session_id

    @property
    def can_subscribe(self) -> bool:
        return self.subscription_count < 300

    @property
    def subscription_count(self) -> int:
        return len(self._subscriptions)

    async def connect(self, *, url: str | None = None, reconnect: bool = False, fail_once: bool = False) -> None:
        if self._closed or self._connecting:
            return

        self._connecting = True
        url_: str = url or f"{WSS}?keepalive_timeout_seconds={self._keep_alive_timeout}"

        self._ready.clear()

        retries: int | None = self._reconnect_attempts
        if retries == 0 and reconnect:
            logger.info("Websocket <%s> was closed unexepectedly, but is flagged as 'should not reconnect'.", self)
            return await self.close()

        while True:
            try:
                async with aiohttp.ClientSession() as session:
                    new = await session.ws_connect(url_, heartbeat=self._heartbeat)
                    session.detach()
            except Exception as e:
                logger.debug("Failed to connect to eventsub websocket <%s>: %s.", self, e)

                if fail_once:
                    await self.close()
                    raise WebsocketConnectionException from e
            else:
                break

            if retries == 0:
                await self.close()

                raise WebsocketConnectionException(
                    "Failed to connect to eventsub websocket <%s> after %s retries. "
                    "Please attempt to reconnect or re-subscribe this eventsub connection.",
                    self,
                    self._reconnect_attempts,
                )

            if retries is not None:
                retries -= 1

            delay: float = self._backoff.calculate()
            logger.info('Websocket <%s> retrying to reconnect websocket connection in "%s" seconds.', self, delay)

            await asyncio.sleep(delay)

        if reconnect:
            await self.close(cleanup=False)

        self._socket = new

        if not self._listen_task:
            self._listen_task = asyncio.create_task(self._listen())

        try:
            async with asyncio.timeout(10 + 1):
                await self._ready.wait()
        except TimeoutError:
            await self.close()

            raise WebsocketConnectionException(
                "Websocket <%s> did not receive a welcome message from Twitch within the allowed timeframe. "
                "Please attempt to reconnect or re-subscribe this eventsub connection.",
                self,
            )

        self._keep_alive_task = asyncio.create_task(self._process_keepalive())

        if reconnect:
            await self._resubscribe()

        self._connecting = False

    async def _resubscribe(self) -> None:
        assert self._session_id

        old_subs = self._subscriptions.copy()
        self._subscriptions.clear()

        for identifier, sub in old_subs.items():
            sub["transport"]["session_id"] = self._session_id

            try:
                resp: SubscriptionResponse = await self._http.create_eventsub_subscription(**sub)
            except HTTPException as e:
                if e.status == 409:
                    # This should never happen here...
                    # But we may as well handle it in-case of edge cases instead of being noisy...

                    msg: str = "Disregarding. Websocket '%s' tried to resubscribe to subscription '%s' but failed with 409."
                    logger.debug(msg, self, identifier)
                    continue

                logger.error("Unable to resubscribe to subscription '%s' on websocket '%s': %s", identifier, self, e)
                continue

            for new in resp["data"]:
                self._subscriptions[new["id"]] = sub

            type_: str = sub["type"].value
            version: str = sub["version"]
            condition: Condition = sub["condition"]

            msg: str = "Websocket '%s' successfully resubscribed to subscription '%s:%s' after reconnect: %s"
            logger.debug(msg, self, type_, version, condition)

    async def _reconnect(self, url: str) -> None:
        socket: Websocket = Websocket(
            keep_alive_timeout=self._keep_alive_timeout,
            reconnect_attempts=self._original_attempts,
            client=self._client,
            token_for=self._token_for,
            http=self._http,
        )

        socket._subscriptions = self._subscriptions

        try:
            await socket.connect(url=url, reconnect=False, fail_once=True)
        except Exception:
            return await self.connect(reconnect=True)

        if self._client:
            self._client._websockets[self._token_for][socket.session_id] = socket  # type: ignore

        await self.close()

    def _create_connection_task(self) -> None:
        task = asyncio.create_task(self.connect(reconnect=True))
        self._connection_tasks.add(task)
        task.add_done_callback(self._connection_tasks.discard)

    async def _listen(self) -> None:
        assert self._socket

        while True:
            try:
                message: aiohttp.WSMessage = await self._socket.receive()
            except Exception:
                self._create_connection_task()
                break

            type_: aiohttp.WSMsgType = message.type
            if type_ in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING):
                logger.debug("Received close message [%s] on eventsub websocket: <%s>", self._socket.close_code, self)

                if self._socket.close_code == 4001:
                    logger.critical(
                        "Websocket <%s> attempted to send an outgoing message to Twitch. "
                        "Twitch prohibits sending outgoing messages to the server, this will result in a disconnect. "
                        "This websocket will NOT attempt to reconnect.",
                        self,
                    )
                    return await self.close()

                elif self._socket.close_code == 4003:
                    return await self.close()

                self._create_connection_task()
                break

            if type_ is not aiohttp.WSMsgType.TEXT:
                logger.debug("Received unknown message from eventsub websocket: <%s>", self)
                continue

            self._last_keepalive = datetime.datetime.now()

            try:
                data: WebsocketMessages = cast("WebsocketMessages", _from_json(message.data))
            except Exception:
                logger.warning("Unable to parse JSON in eventsub websocket: <%s>", self)
                continue

            metadata: MetaData = data["metadata"]
            message_type: MessageTypes = metadata["message_type"]

            if message_type == "session_welcome":
                welcome_data: WelcomeMessage = cast("WelcomeMessage", data)

                await self._process_welcome(welcome_data)

            elif message_type == "session_reconnect":
                logger.debug('Received "session_reconnect" message from eventsub websocket: <%s>', self)
                reconnect_data: ReconnectMessage = cast("ReconnectMessage", data)

                await self._process_reconnect(reconnect_data)

            elif message_type == "session_keepalive":
                logger.debug('Received "session_keepalive" message from eventsub websocket: <%s>', self)

            elif message_type == "revocation":
                logger.debug('Received "revocation" message from eventsub websocket: <%s>', self)

                revocation_data: RevocationMessage = cast("RevocationMessage", data)
                await self._process_revocation(revocation_data)

            elif message_type == "notification":
                logger.debug('Received "notification" message from eventsub websocket: <%s>. %s', self, data)
                notification_data: NotificationMessage = cast("NotificationMessage", data)

                try:
                    await self._process_notification(notification_data)
                except Exception as e:
                    msg = "Caught an unknown exception while proccessing a websocket 'notification' event:\n%s\n"
                    logger.critical(msg, str(e), exc_info=e)

            else:
                logger.warning("Received an unknown message type in eventsub websocket: <%s>", self)

    async def _process_keepalive(self) -> None:
        assert self._last_keepalive
        logger.debug("Started keep_alive task on eventsub websocket: <%s>", self)

        while True:
            await asyncio.sleep(self._keep_alive_timeout)
            now: datetime.datetime = datetime.datetime.now()

            if self._last_keepalive + datetime.timedelta(seconds=self._keep_alive_timeout + 5) < now:
                self._create_connection_task()
                return

    async def _process_welcome(self, data: WelcomeMessage) -> None:
        payload: WelcomePayload = data["payload"]
        new_id: str = payload["session"]["id"]

        if self._session_id:
            self._cleanup(closed=False)
            self._client._websockets[self._token_for] = {self.session_id: self}  # type: ignore

        self._session_id = new_id
        self._ready.set()

        assert self._listen_task

        self._listen_task.set_name(f"EventsubWebsocketListener: {self._session_id}")
        logger.info('Received "session_welcome" message from eventsub websocket: <%s>', self)

    async def _process_reconnect(self, data: ReconnectMessage) -> None:
        logger.info("Attempting to reconnect eventsub websocket due to a reconnect message from Twitch: <%s>", self)
        await self._reconnect(url=data["payload"]["session"]["reconnect_url"])

    async def _process_revocation(self, data: RevocationMessage) -> None:
        payload: SubscriptionRevoked = SubscriptionRevoked(data=data["payload"]["subscription"])

        if self._client:
            self._client.dispatch(event="subscription_revoked", payload=payload)

        self._subscriptions.pop(payload.id, None)
        if not self._subscriptions:
            logger.info("Closing websocket '%s' due to no remaining subscriptions.", self)
            return await self.close()

    async def _process_notification(self, data: NotificationMessage) -> None:
        sub_type = data["metadata"]["subscription_type"]
        event = _SUB_MAPPING.get(sub_type, sub_type.removeprefix("channel.")).replace(".", "_")

        try:
            payload_class = create_event_instance(sub_type, data, http=self._http)
        except ValueError:
            logger.warning("Websocket '%s' received an unhandled eventsub event: '%s'.", self, event)
            return

        if self._client:
            self._client.dispatch(event=event, payload=payload_class)

    def _cleanup(self, closed: bool = True) -> None:
        self._closed = closed

        if self._client:
            sockets = self._client._websockets.get(self._token_for, {})
            sockets.pop(self.session_id or "", None)

    async def close(self, cleanup: bool = True) -> None:
        self._closing = True

        if cleanup:
            self._cleanup()

        if self._keep_alive_task:
            try:
                self._keep_alive_task.cancel()
            except Exception:
                pass

        if self._listen_task:
            try:
                self._listen_task.cancel()
            except Exception:
                pass

        if self._socket:
            try:
                await self._socket.close()
            except Exception:
                pass

        self._keep_alive_task = None
        self._listen_task = None
        self._socket = None

        logger.info("Successfully closed eventsub websocket: <%s>", self)
        self._closing = False



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import io
import logging
import pathlib
from typing import TYPE_CHECKING, Any

import yarl

from .exceptions import HTTPException


if TYPE_CHECKING:
    import os

    from .http import HTTPClient


logger: logging.Logger = logging.getLogger(__name__)


VALID_ASSET_EXTENSIONS: set[str] = {
    ".png",
    ".jpg",
    ".jpeg",
    ".gif",
    ".webp",
}


class Asset:
    """Represents an asset from Twitch.

    Assets can be used to save or read from images or other media from Twitch.
    You can also retrieve the URL of the asset via the provided properties and methods.

    .. versionadded:: 3.0.0
        Added the asset class which will replace all
        previous properties of models with attached media URLs.

    Supported Operations
    --------------------

    +-------------+-------------------------------------------+-----------------------------------------------+
    | Operation   | Usage(s)                                  | Description                                   |
    +=============+===========================================+===============================================+
    | __str__     | ``str(asset)``, ``f"{asset}"``            | Returns the asset's URL.                      |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | __repr__    | ``repr(asset)``, ``f"{asset!r}"``         | Returns the asset's official representation.  |
    +-------------+-------------------------------------------+-----------------------------------------------+
    """

    __slots__ = ("_dimensions", "_ext", "_http", "_name", "_original_url", "_url")

    def __init__(
        self,
        url: str,
        *,
        http: HTTPClient,
        name: str | None = None,
        dimensions: tuple[int, int] | None = None,
    ) -> None:
        self._http: HTTPClient = http

        ext: str = yarl.URL(url).suffix
        self._ext: str | None = ext if ext in VALID_ASSET_EXTENSIONS else None

        self._dimensions: tuple[int, int] | None = dimensions
        self._original_url: str = url
        self._url: str = url.format(width=dimensions[0], height=dimensions[1]) if dimensions else url
        self._name: str = name or yarl.URL(self._url).name

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"Asset(name={self.name}, url={self.url})"

    @property
    def url(self) -> str:
        """The URL of the asset.

        If the asset supports custom dimensions, the URL will contain the dimensions set.

        See :meth:`.set_dimensions` for information on setting custom dimensions.
        """
        return self._url

    @property
    def base_url(self) -> str:
        """The base URL of the asset without any dimensions set.

        This is the URL provided by Twitch before any dimensions are set.
        """
        return self._original_url

    @property
    def name(self) -> str:
        """A property that returns the default name of the asset."""
        return self._name

    @property
    def qualified_name(self) -> str:
        """A property that returns the qualified name of the asset.

        This is the name of the asset with the file extension if one can be determined.
        If the file extension has not been set, this method returns the same as :attr:`.name`.
        """
        name: str = self._name.split(".")[0]
        return f"{name}{self._ext}" if self._ext else self._name

    @property
    def ext(self) -> str | None:
        """A property that returns the file extension of the asset.

        Could be ``None`` if the asset does not have a valid file extension or it has not been determined yet.

        See: `:meth:`.fetch_ext` to try and force setting the file extension by content type.
        """
        return self._ext.removeprefix(".") if self._ext else None

    @property
    def dimensions(self) -> tuple[int, int] | None:
        """A property that returns the dimensions of the asset if it supports custom dimensions or ``None``.

        See: :meth:`.set_dimensions` for more information.
        """
        return self._dimensions

    def set_dimensions(self, width: int, height: int) -> None:
        """Set the dimensions of the asset for saving or reading.

        By default all assets that support custom dimensions already have pre-defined values set.
        If custom dimensions are **not** supported, a warning will be logged and the default dimensions will be used.

        .. warning::
            If you need to custom dimensions for an asset that supports it you should use this method **before**
            calling :meth:`.save` or :meth:`.read`.

        Examples
        --------
        .. code:: python3

            # Fetch a game and set the box art dimensions to 720x960; which is a 3:4 aspect ratio.

            game: twitchio.Game = await client.fetch_game("League of Legends")
            game.box_art.set_dimensions(720, 960)

            # Call read or save...
            await game.box_art.save()


        Parameters
        ----------
        width: int
            The width of the asset.
        height: int
            The height of the asset.
        """
        if not self._dimensions:
            logger.warning("Setting dimensions on asset %r is not supported.", self)
            return

        self._dimensions = (width, height)
        self._url = self._original_url.format(width=width, height=height)

    def url_for(self, width: int, height: int) -> str:
        """Return a new URL for the asset with the specified dimensions.

        .. note::
            This method does not return new dimensions on assets that do not support it.

        .. warning::
            This method does not set dimensions for saving or reading.
            If you need custom dimensions for an asset that supports it see: :meth:`.set_dimensions`.

        Parameters
        ----------
        width: int
            The width of the asset.
        height: int
            The height of the asset.

        Returns
        -------
        str
            The new URL for the asset with the specified dimensions or
            the original URL if the asset does not support custom dimensions.
        """
        if not self._dimensions:
            logger.warning("Setting dimensions on asset %r is not supported.", self)
            return self._url

        return self._original_url.format(width=width, height=height)

    def _set_ext(self, headers: dict[str, str]) -> str | None:
        content: str | None = headers.get("Content-Type")
        if not content or not content.startswith("image/"):
            return None

        ext: str = content.split("/")[1]
        self._ext = f".{ext}"

        return self._ext

    async def fetch_ext(self) -> str | None:
        """Fetch and set the file extension of the asset by content type.

        This method will try to fetch the file extension of the asset by making a HEAD request to the asset's URL.
        If the content type is not recognized or the request fails, the file extension will remain unchanged.

        For the majority of cases you should not need to use this method.

        .. warning::
            This method sets the file extension of the asset by content type.

        Returns
        -------
        str | None
            The file extension of the asset determined by the content type or ``None`` if it could not be determined.
        """
        try:
            headers: dict[str, str] = await self._http._request_asset_head(self.url)
        except HTTPException:
            return None

        return self._set_ext(headers)

    async def save(
        self,
        fp: str | os.PathLike[Any] | io.BufferedIOBase | None = None,
        seek_start: bool = True,
        force_extension: bool = True,
    ) -> int:
        """Save this asset to a file or file-like object.

        If ``fp`` is ``None``, the asset will be saved to the current working directory with the
        asset's default qualified name.

        Examples
        --------

        **Save with defaults**

            .. code:: python3

                # Fetch a game and save the box art to the current working directory with the asset's default name.

                game: twitchio.Game = await client.fetch_game("League of Legends")
                await game.box_art.save()


        **Save with a custom name**

            .. code:: python3

                # Fetch a game and save the box art to the current working directory with a custom name.

                game: twitchio.Game = await client.fetch_game("League of Legends")
                await game.box_art.save("custom_name.png")


        **Save with a file-like object**

            .. code:: python3

                # Fetch a game and save the box art to a file-like object.

                game: twitchio.Game = await client.fetch_game("League of Legends")
                with open("custom_name.png", "wb") as fp:
                    await game.box_art.save(fp)


        Parameters
        -----------
        fp: str | os.PathLike | io.BufferedIOBase | None
            The file path or file-like object to save the asset to.

            If ``None``, the asset will be saved to the current working directory with the asset's qualified name.

            If ``fp`` is a directory, the asset will be saved to the directory with the asset's qualified name.

            Defaults to ``None``.
        seek_start: bool
            Whether to seek to the start of the file after successfully writing data. Defaults to ``True``.
        force_extension: bool
            Whether to force the file extension of the asset to match the content type. Defaults to ``True``.

            If no file extension was provided with ``fp`` setting ``force_extension`` to ``True``
            will force the file extension to match the content type provided by Twitch.

        Returns
        -------
        int
            The number of bytes written to the file or file-like object.

        Raises
        ------
        FileNotFoundError
            Raised when ``fp`` is a directory or path to directory which can not be found or accessed.
        """
        data: io.BytesIO = await self.read()
        written: int = 0

        if isinstance(fp, io.BufferedIOBase):
            written = fp.write(data.read())

            if seek_start:
                fp.seek(0)

            return written

        if not fp:
            fp = pathlib.Path.cwd() / self.qualified_name

        elif pathlib.Path(fp).is_dir():
            fp = pathlib.Path(fp) / (self.qualified_name if force_extension else self.name)

        elif isinstance(fp, str) and force_extension:
            fp = f"{fp}{self._ext or ''}"

        with open(fp, "wb") as new:
            written = new.write(data.read())

        return written

    async def read(self, *, seek_start: bool = True, chunk_size: int = 1024) -> io.BytesIO:
        """Read from the asset and return an :class:`io.BytesIO` buffer.

        You can use this method to save the asset to memory and use it later.

        Examples
        --------
        .. code:: python3

            # Fetch a game and read the box art to memory.

            game: twitchio.Game = await client.fetch_game("League of Legends")
            data: io.BytesIO = await game.box_art.read()

            # Later...
            some_bytes = data.read()


        Parameters
        ----------
        seek_start: bool
            Whether to seek to the start of the buffer after successfully writing data. Defaults to ``True``.
        chunk_size: int
            The size of the chunk to use when reading from the asset. Defaults to ``1024``.

        Returns
        -------
        io.BytesIO
            A bytes buffer containing the asset's data.
        """
        fp: io.BytesIO = io.BytesIO()

        async for chunk in self._http._request_asset(self, chunk_size=chunk_size):
            fp.write(chunk)

        if seek_start:
            fp.seek(0)

        return fp



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from twitchio.assets import Asset
from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp

from .games import Game


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import GamesResponse, SearchChannelsResponseData


__all__ = ("SearchChannel",)


class SearchChannel:
    """Represents a channel via a search.

    Attributes
    -----------
    broadcaster: PartialUser
        The broadcaster / channel.
    title: str
        The stream's title. Is an empty string if the broadcaster didn't set it.
    game_id: str
        The ID of the game that the broadcaster is playing or last played.
    game_name: str
        The name of the game that the broadcaster is playing or last played.
    live: bool
        A Boolean value that determines whether the broadcaster is streaming live. Is True if the broadcaster is streaming live; otherwise, False.
    tags: list[str]
        The tags applied to the channel.
    thumbnail: Asset
        An Asset for the thumbnail of the broadcaster's profile image.
    language: str
        The ISO 639-1 two-letter language code of the language used by the broadcaster. For example, en for English. If the broadcaster uses a language not in the list of `supported stream languages <https://help.twitch.tv/s/article/languages-on-twitch#streamlang>`_, the value is other.
    started_at: datetime.datetime | None
        Datetime of when the broadcaster started streaming. Is None if the broadcaster is not streaming live.

    """

    __slots__ = (
        "_http",
        "broadcaster",
        "game_id",
        "language",
        "live",
        "name",
        "started_at",
        "tags",
        "thumbnail",
        "title",
    )

    def __init__(self, data: SearchChannelsResponseData, *, http: HTTPClient) -> None:
        self._http: HTTPClient = http
        self.broadcaster: PartialUser = PartialUser(data["id"], data["broadcaster_login"], data["display_name"], http=http)
        self.game_id: str = data["game_id"]
        self.title: str = data["title"]
        self.thumbnail: Asset = Asset(data["thumbnail_url"], http=http)
        self.language: str = data["broadcaster_language"]
        self.live: bool = data["is_live"]
        self.started_at: datetime.datetime | None = parse_timestamp(data["started_at"]) if self.live else None
        self.tags: list[str] = data["tags"]

    def __repr__(self) -> str:
        return f"<SearchChannel broadcaster={self.broadcaster} title={self.title} live={self.live} game_id={self.game_id}>"

    async def fetch_game(self) -> Game | None:
        """|coro|

        Fetches the :class:`~twitchio.Game` associated with this channel.

        The :class:`~twitchio.Game` returned is current from the time the :class:`~twitchio.SearchChannel`
        instance was created.

        Returns
        -------
        twitchio.Game
            The game associated with this :class:`~twitchio.SearchChannel` instance.
        """
        payload: GamesResponse = await self._http.get_games(ids=[self.game_id])
        return Game(payload["data"][0], http=self._http) if payload["data"] else None



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import enum


__all__ = (
    "CloseCode",
    "MessageType",
    "RevocationReason",
    "ShardStatus",
    "SubscriptionType",
    "TransportMethod",
)


class TransportMethod(enum.Enum):
    """An enum of EventSub transports.

    Attributes
    ----------
    WEBHOOK: Literal["webhook"]
    WEBSOCKET: Literal["websocket]
    CONDUIT: Literal["conduit"]
    """

    WEBHOOK = "webhook"
    WEBSOCKET = "websocket"
    CONDUIT = "conduit"


class ShardStatus(enum.Enum):
    ENABLED = "enabled"
    WEBHOOK_VERIFICATION_PENDING = "webhook_callback_verification_pending"
    WEBHOOK_VERIFICATION_FAILED = "webhook_callback_verification_failed"
    NOTIFICATION_FAILURES_EXCEEDED = "notification_failures_exceeded"
    WEBSOCKET_DISCONNECTED = "websocket_disconnected"
    WEBSOCKET_FAILIED_PING = "websocket_failed_ping_pong"
    WEBSOCKET_RECEIVED_INBOUD = "websocket_received_inbound_traffic"
    WEBSOCKET_INTERNAL_ERROR = "websocket_internal_error"
    WEBSOCKET_NETWORK_TIMEOUT = "websocket_network_timeout"
    WEBSOCKET_NETWORK_ERROR = "websocket_network_error"
    WEBSOCKET_FAILED_RECONNECT = "websocket_failed_to_reconnect"


class CloseCode(enum.Enum):
    INTERNAL_SERVER_ERROR = 4000
    SENT_INBOUND_TRAFFIC = 4001
    FAILED_PING = 4002
    CONNECTION_UNUSED = 4003
    RECONNECT_GRACE_TIMEOUT = 4004
    NETWORK_TIMEOUT = 4005
    NETWORK_ERROR = 4006
    INVALID_RECONNECT = 4007


class MessageType(enum.Enum):
    SESSION_WELCOME = "session_welcome"
    SESSION_KEEPALIVE = "session_keepalive"
    NOTIFICATION = "notification"
    SESSION_RECONNECT = "session_reconnect"
    REVOCATION = "revocation"


class SubscriptionType(enum.Enum):
    """An enum of available EventSub subscription types.

    The value of each enum attribute will be the formal name sent and received to/from Twitch.

    E.g. ``AutomodMessageHold`` -> ``"automod.message.hold"``.

    Attributes
    ----------
    AutomodMessageHold: Literal["automod.message.hold"]
    AutomodMessageUpdate: Literal["automod.message.update"]
    AutomodSettingsUpdate: Literal["automod.settings.update"]
    AutomodTermsUpdate: Literal["automod.terms.update"]
    ChannelUpdate: Literal["channel.update"]
    ChannelFollow: Literal["channel.follow"]
    ChannelAdBreakBegin: Literal["channel.ad_break.begin"]
    ChannelBitsUseSubscription: Literal["channel.bits.use"]
    ChannelChatClear: Literal["channel.chat.clear"]
    ChannelChatClearUserMessages: Literal["channel.chat.clear_user_messages"]
    ChannelChatMessage: Literal["channel.chat.message"]
    ChannelChatMessageDelete: Literal["channel.chat.message_delete"]
    ChannelChatNotification: Literal["channel.chat.notification"]
    ChannelChatSettingsUpdate: Literal["channel.chat_settings.update"]
    ChannelChatUserMessageHold: Literal["channel.chat.user_message_hold"]
    ChannelChatUserMessageUpdate: Literal["channel.chat.user_message_update"]
    ChannelSharedChatBegin: Literal["channel.shared_chat.begin"]
    ChannelSharedChatUpdate: Literal["channel.shared_chat.update"]
    ChannelSharedChatEnd: Literal["channel.shared_chat.end"]
    ChannelSubscribe: Literal["channel.subscribe"]
    ChannelSubscriptionEnd: Literal["channel.subscription.end"]
    ChannelSubscriptionGift: Literal["channel.subscription.gift"]
    ChannelSubscriptionMessage: Literal["channel.subscription.message"]
    ChannelCheer: Literal["channel.cheer"]
    ChannelRaid: Literal["channel.raid"]
    ChannelBan: Literal["channel.ban"]
    ChannelUnban: Literal["channel.unban"]
    ChannelUnbanRequestCreate: Literal["channel.unban_request.create"]
    ChannelUnbanRequestResolve: Literal["channel.unban_request.resolve"]
    ChannelModerate: Literal["channel.moderate"]
    ChannelModeratorAdd: Literal["channel.moderator.add"]
    ChannelModeratorRemove: Literal["channel.moderator.remove"]
    ChannelGuestStarSessionBegin: Literal["channel.guest_star_session.begin"]
    ChannelGuestStarSessionEnd: Literal["channel.guest_star_session.end"]
    ChannelGuestStarGuestUpdate: Literal["channel.guest_star_guest.update"]
    ChannelGuestStarSettingsUpdate: Literal["channel.guest_star_settings.update"]
    ChannelChannelPointsAutomaticRewardRedemptionAdd: Literal["channel.channel_points_automatic_reward_redemption.add"]
    ChannelChannelPointsCustomRewardAdd: Literal["channel.channel_points_custom_reward.add"]
    ChannelChannelPointsCustomRewardUpdate: Literal["channel.channel_points_custom_reward.update"]
    ChannelChannelPointsCustomRewardRemove: Literal["channel.channel_points_custom_reward.remove"]
    ChannelChannelPointsCustomRewardRedemptionAdd: Literal["channel.channel_points_custom_reward_redemption.add"]
    ChannelChannelPointsCustomRewardRedemptionUpdate: Literal["channel.channel_points_custom_reward_redemption.update"]
    ChannelPollBegin: Literal["channel.poll.begin"]
    ChannelPollProgres: Literal["channel.poll.progress"]
    ChannelPollEnd: Literal["channel.poll.end"]
    ChannelPredictionBegin: Literal["channel.prediction.begin"]
    ChannelPredictionProgres: Literal["channel.prediction.progress"]
    ChannelPredictionLock: Literal["channel.prediction.lock"]
    ChannelPredictionEnd: Literal["channel.prediction.end"]
    ChannelSuspiciousUserMessage: Literal["channel.suspicious_user.message"]
    ChannelSuspiciousUserUpdate: Literal["channel.suspicious_user.update"]
    ChannelVipAdd: Literal["channel.vip.add"]
    ChannelVipRemove: Literal["channel.vip.remove"]
    ChannelCharityCampaignDonate: Literal["channel.charity_campaign.donate"]
    ChannelCharityCampaignStart: Literal["channel.charity_campaign.start"]
    ChannelCharityCampaignProgress: Literal["channel.charity_campaign.progress"]
    ChannelCharityCampaignStop: Literal["channel.charity_campaign.stop"]
    ConduitShardDisabled: Literal["conduit.shard.disabled"]
    DropEntitlementGrant: Literal["drop.entitlement.grant"]
    ExtensionBitsTransactionCreate: Literal["extension.bits_transaction.create"]
    ChannelGoalBegin: Literal["channel.goal.begin"]
    ChannelGoalProgress: Literal["channel.goal.progress"]
    ChannelGoalEnd: Literal["channel.goal.end"]
    ChannelHypeTrainBegin: Literal["channel.hype_train.begin"]
    ChannelHypeTrainProgress: Literal["channel.hype_train.progress"]
    ChannelHypeTrainEnd: Literal["channel.hype_train.end"]
    ChannelShieldModeBegin: Literal["channel.shield_mode.begin"]
    ChannelShieldModeEnd: Literal["channel.shield_mode.end"]
    ChannelShoutoutCreate: Literal["channel.shoutout.create"]
    ChannelShoutoutReceive: Literal["channel.shoutout.receive"]
    ChannelWarningAcknowledgement: Literal["channel.warning.acknowledge"]
    ChannelWarningSend: Literal["channel.warning.send"]
    StreamOnline: Literal["stream.online"]
    StreamOffline: Literal["stream.offline"]
    UserAuthorizationGrant: Literal["user.authorization.grant"]
    UserAuthorizationRevoke: Literal["user.authorization.revoke"]
    UserUpdate: Literal["user.update"]
    UserWhisperMessage: Literal["user.whisper.message"]
    """

    AutomodMessageHold = "automod.message.hold"
    AutomodMessageUpdate = "automod.message.update"
    AutomodSettingsUpdate = "automod.settings.update"
    AutomodTermsUpdate = "automod.terms.update"
    ChannelBitsUse = "channel.bits.use"
    ChannelUpdate = "channel.update"
    ChannelFollow = "channel.follow"
    ChannelAdBreakBegin = "channel.ad_break.begin"
    ChannelChatClear = "channel.chat.clear"
    ChannelChatClearUserMessages = "channel.chat.clear_user_messages"
    ChannelChatMessage = "channel.chat.message"
    ChannelChatMessageDelete = "channel.chat.message_delete"
    ChannelChatNotification = "channel.chat.notification"
    ChannelChatSettingsUpdate = "channel.chat_settings.update"
    ChannelChatUserMessageHold = "channel.chat.user_message_hold"
    ChannelChatUserMessageUpdate = "channel.chat.user_message_update"
    ChannelSharedChatBegin = "channel.shared_chat.begin"
    ChannelSharedChatUpdate = "channel.shared_chat.update"
    ChannelSharedChatEnd = "channel.shared_chat.end"
    ChannelSubscribe = "channel.subscribe"
    ChannelSubscriptionEnd = "channel.subscription.end"
    ChannelSubscriptionGift = "channel.subscription.gift"
    ChannelSubscriptionMessage = "channel.subscription.message"
    ChannelCheer = "channel.cheer"
    ChannelRaid = "channel.raid"
    ChannelBan = "channel.ban"
    ChannelUnban = "channel.unban"
    ChannelUnbanRequestCreate = "channel.unban_request.create"
    ChannelUnbanRequestResolve = "channel.unban_request.resolve"
    ChannelModerate = "channel.moderate"
    ChannelModeratorAdd = "channel.moderator.add"
    ChannelModeratorRemove = "channel.moderator.remove"
    ChannelGuestStarSessionBegin = "channel.guest_star_session.begin"
    ChannelGuestStarSessionEnd = "channel.guest_star_session.end"
    ChannelGuestStarGuestUpdate = "channel.guest_star_guest.update"
    ChannelGuestStarSettingsUpdate = "channel.guest_star_settings.update"
    ChannelChannelPointsAutomaticRewardRedemptionAdd = "channel.channel_points_automatic_reward_redemption.add"
    ChannelChannelPointsCustomRewardAdd = "channel.channel_points_custom_reward.add"
    ChannelChannelPointsCustomRewardUpdate = "channel.channel_points_custom_reward.update"
    ChannelChannelPointsCustomRewardRemove = "channel.channel_points_custom_reward.remove"
    ChannelChannelPointsCustomRewardRedemptionAdd = "channel.channel_points_custom_reward_redemption.add"
    ChannelChannelPointsCustomRewardRedemptionUpdate = "channel.channel_points_custom_reward_redemption.update"
    ChannelPollBegin = "channel.poll.begin"
    ChannelPollProgress = "channel.poll.progress"
    ChannelPollEnd = "channel.poll.end"
    ChannelPredictionBegin = "channel.prediction.begin"
    ChannelPredictionProgress = "channel.prediction.progress"
    ChannelPredictionLock = "channel.prediction.lock"
    ChannelPredictionEnd = "channel.prediction.end"
    ChannelSuspiciousUserMessage = "channel.suspicious_user.message"
    ChannelSuspiciousUserUpdate = "channel.suspicious_user.update"
    ChannelVipAdd = "channel.vip.add"
    ChannelVipRemove = "channel.vip.remove"
    ChannelCharityCampaignDonate = "channel.charity_campaign.donate"
    ChannelCharityCampaignStart = "channel.charity_campaign.start"
    ChannelCharityCampaignProgress = "channel.charity_campaign.progress"
    ChannelCharityCampaignStop = "channel.charity_campaign.stop"
    ConduitShardDisabled = "conduit.shard.disabled"
    DropEntitlementGrant = "drop.entitlement.grant"
    ExtensionBitsTransactionCreate = "extension.bits_transaction.create"
    ChannelGoalBegin = "channel.goal.begin"
    ChannelGoalProgress = "channel.goal.progress"
    ChannelGoalEnd = "channel.goal.end"
    ChannelHypeTrainBegin = "channel.hype_train.begin"
    ChannelHypeTrainProgress = "channel.hype_train.progress"
    ChannelHypeTrainEnd = "channel.hype_train.end"
    ChannelShieldModeBegin = "channel.shield_mode.begin"
    ChannelShieldModeEnd = "channel.shield_mode.end"
    ChannelShoutoutCreate = "channel.shoutout.create"
    ChannelShoutoutReceive = "channel.shoutout.receive"
    ChannelWarningAcknowledgement = "channel.warning.acknowledge"
    ChannelWarningSend = "channel.warning.send"
    StreamOnline = "stream.online"
    StreamOffline = "stream.offline"
    UserAuthorizationGrant = "user.authorization.grant"
    UserAuthorizationRevoke = "user.authorization.revoke"
    UserUpdate = "user.update"
    UserWhisperMessage = "user.whisper.message"


class RevocationReason(enum.Enum):
    USER_REMOVED = "user_removed"
    AUTHORIZATION_REVOKED = "authorization_revoked"
    NOTIFICATION_FAILURES_EXCEEDED = "notification_failures_exceeded"
    VERSION_REMOVED = "version_removed"
